@@ -1,6 +1,6 @@
 /* libpoke.h - Public header for libpoke.  */
 
-/* Copyright (C) 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +29,13 @@
 #define LIBPOKE_API
 #endif
 
+#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 3) > 3)                \
+    || (defined __clang__ && __clang_major__ >= 3)
+#define LIBPOKE_NONNULL(...) __attribute__ ((nonnull (__VA_ARGS__)))
+#else
+#define LIBPOKE_NONNULL(...)
+#endif
+
 typedef struct _pk_compiler *pk_compiler;
 typedef struct _pk_ios *pk_ios;
 typedef uint64_t pk_val;
@@ -49,7 +56,6 @@ typedef uint64_t pk_val;
 #define PK_EC_NO_IOS        2
 #define PK_EC_NO_RETURN     3
 #define PK_EC_OUT_OF_BOUNDS 4
-#define PK_EC_MAP_BOUNDS    5
 #define PK_EC_EOF           6
 #define PK_EC_MAP           7
 #define PK_EC_CONV          8
@@ -63,6 +69,7 @@ typedef uint64_t pk_val;
 #define PK_EC_ASSERT       16
 #define PK_EC_OVERFLOW     17
 #define PK_EC_PERM         18
+#define PK_EC_STACK        19
 
 struct pk_color
 {
@@ -76,53 +83,50 @@ struct pk_color
 struct pk_term_if
 {
   /* Flush the output in the terminal.  */
-  void (*flush_fn) (void);
+  void (*flush_fn) (pk_compiler pkc);
 
   /* Output a NULL-terminated C string.  */
-  void (*puts_fn) (const char *str);
+  void (*puts_fn) (pk_compiler pkc, const char *str);
 
   /* Output a formatted string.  */
-  void (*printf_fn) (const char *format, ...);
+  void (*printf_fn) (pk_compiler pkc, const char *format, ...);
 
   /* Output LVL levels of indentation, using STEP white chars in each
      indentation level.  */
-  void (*indent_fn) (unsigned int lvl, unsigned int step);
+  void (*indent_fn) (pk_compiler pkc, unsigned int lvl, unsigned int step);
 
   /* Mark the beginning of a styling class with name CLASS.  */
-  void (*class_fn) (const char *aclass);
+  void (*class_fn) (pk_compiler pkc, const char *aclass);
 
   /* Mark the end of a styling class with name CLASS.  This function
      returns 0 if the given class is not active and therefore can't be
      ended.  1 otherwise.  */
-  int (*end_class_fn) (const char *aclass);
+  int (*end_class_fn) (pk_compiler pkc, const char *aclass);
 
   /* Mark the beginning of an hyperlink with url URL and identifier
      ID.  The identifier can be NULL.  */
-  void (*hyperlink_fn) (const char *url, const char *id);
+  void (*hyperlink_fn) (pk_compiler pkc, const char *url, const char *id);
 
   /* Mark the end of the current hyperlink.  This function returns 0
      if there is no currently an hyperlink open to close.  1
      otherwise.  */
-  int (*end_hyperlink_fn) (void);
+  int (*end_hyperlink_fn) (pk_compiler pkc);
 
   /* Get the current foreground color.  */
-  struct pk_color (*get_color_fn) (void);
+  struct pk_color (*get_color_fn) (pk_compiler pkc);
 
   /* Get the current background color.  */
-  struct pk_color (*get_bgcolor_fn) (void);
+  struct pk_color (*get_bgcolor_fn) (pk_compiler pkc);
 
   /* Set the foreground color.  */
-  void (*set_color_fn) (struct pk_color color);
+  void (*set_color_fn) (pk_compiler pkc, struct pk_color color);
 
   /* Set the background color.  */
-  void (*set_bgcolor_fn) (struct pk_color color);
+  void (*set_bgcolor_fn) (pk_compiler pkc, struct pk_color color);
 };
 
 /* Create and return a new Poke incremental compiler.
 
-   RTPATH should contain the name of a directory where the compiler
-   can find its run-time support files.
-
    TERM_IF is a pointer to a struct pk_term_if containing pointers to
    functions providing the output routines to be used by the
    incremental compiler.
@@ -146,6 +150,13 @@ pk_compiler pk_compiler_new_with_flags (struct pk_term_if *term_if,
 
 void pk_compiler_free (pk_compiler pkc) LIBPOKE_API;
 
+/* Register/unregister a new thread (other than the thread that called
+   pk_compiler_new) that calls any of the services provided by the
+   library.  */
+
+void pk_register_thread (void) LIBPOKE_API;
+void pk_unregister_thread (void) LIBPOKE_API;
+
 /* Error code of last operation.
 
    This function returns the status corresponding to the given PK
@@ -155,19 +166,27 @@ void pk_compiler_free (pk_compiler pkc) LIBPOKE_API;
 
 int pk_errno (pk_compiler pkc) LIBPOKE_API;
 
-/* Compile an execute a Poke program from the given file FILENAME.
+/* Set/get user-defined opaque data to the compiler instance.
+   The pointer is accessible whenever an instance of a Poke incremental
+   compiler is available (including from callbacks).  */
+
+void pk_set_user_data (pk_compiler pkc, void *user_data) LIBPOKE_API;
+void *pk_get_user_data (pk_compiler pkc) LIBPOKE_API;
+
+/* Compile and execute a Poke program from the given file FILENAME.
 
    EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
-   Exception value if the execution rsults in an unhandled exception,
+   Exception value if the execution results in an unhandled exception,
    PK_NULL otherwise.
 
    Return PK_ERROR in case of a compilation error.  Otherwise,
    return PK_OK.  */
 
 int pk_compile_file (pk_compiler pkc, const char *filename,
-                     pk_val *exit_exception) LIBPOKE_API;
+                     pk_val *exit_exception)
+    LIBPOKE_API LIBPOKE_NONNULL (1, 2, 3);
 
-/* Compile an execute a Poke program from a memory buffer.
+/* Compile and execute a Poke program from a memory buffer.
 
    BUFFER is a NULL-terminated string.
 
@@ -178,11 +197,27 @@ int pk_compile_file (pk_compiler pkc, const char *filename,
    return PK_OK.
 
    EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
-   Exception value if the execution rsults in an unhandled exception,
+   Exception value if the execution results in an unhandled exception,
    PK_NULL otherwise.  */
 
 int pk_compile_buffer (pk_compiler pkc, const char *buffer,
-                       const char **end, pk_val *exit_exception) LIBPOKE_API;
+                       const char **end, pk_val *exit_exception)
+    LIBPOKE_API LIBPOKE_NONNULL (1, 2, 4);
+
+/* Like pk_compile_buffer, but gets the following extra arguments.
+
+   SOURCE is a string identifying the source of the code.  This will
+   often be a file name, or a string like "<stdin>".
+
+   LINE and COLUMN specify the relative location of BUFFER in SOURCE.
+   Both lines and columns start numbering at one, i.e. the first
+   column/line is the column/line number 1.  */
+
+int pk_compile_buffer_with_loc (pk_compiler pkc, const char *buffer,
+                                const char *source,
+                                uint32_t line, uint32_t column,
+                                const char **end, pk_val *exit_exception)
+  LIBPOKE_API LIBPOKE_NONNULL (1, 2, 7);
 
 /* Like pk_compile_buffer but compile and execute a single Poke
    statement, which may evaluate to a value if it is an "expression
@@ -192,12 +227,29 @@ int pk_compile_buffer (pk_compiler pkc, const char *buffer,
    result value of an expression-statement, or to PK_NULL.
 
    EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
-   Exception value if the execution rsults in an unhandled exception,
+   Exception value if the execution results in an unhandled exception,
    PK_NULL otherwise.  */
 
 int pk_compile_statement (pk_compiler pkc, const char *buffer,
                           const char **end, pk_val *val,
-                          pk_val *exit_exception) LIBPOKE_API;
+                          pk_val *exit_exception)
+    LIBPOKE_API LIBPOKE_NONNULL (1, 2, 5);
+
+/* Like pk_compile_statement, but gets the following extra arguments.
+
+   SOURCE is a string identifying the source of the code.  This will
+   often be a file name, or a string like "<stdin>".
+
+   LINE and COLUMN specify the relative location of BUFFER in SOURCE.
+   Both lines and columns start numbering at one, i.e. the first
+   column/line is the column/line number 1.  */
+
+int pk_compile_statement_with_loc (pk_compiler pkc, const char *buffer,
+                                   const char *source,
+                                   uint32_t line, uint32_t column,
+                                   const char **end, pk_val *val,
+                                   pk_val *exit_exception)
+    LIBPOKE_API LIBPOKE_NONNULL (1, 2, 8);
 
 /* Like pk_compile_buffer but compile and execute a single Poke
    expression, which evaluates to a value.
@@ -208,19 +260,45 @@ int pk_compile_statement (pk_compiler pkc, const char *buffer,
    Return PK_ERROR in case of a compilation error, PK_OK otherwise.
 
    EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
-   Exception value if the execution rsults in an unhandled exception,
+   Exception value if the execution results in an unhandled exception,
    PK_NULL otherwise.  */
 
 int pk_compile_expression (pk_compiler pkc, const char *buffer,
                            const char **end, pk_val *val,
-                           pk_val *exit_exception) LIBPOKE_API;
+                           pk_val *exit_exception)
+    LIBPOKE_API LIBPOKE_NONNULL (1, 2, 5);
+
+/* Like pk_compile_expression, but gets the following extra arguments.
+
+   SOURCE is a string identifying the source of the code.  This will
+   often be a file name, or a string like "<stdin>".
+
+   LINE and COLUMN specify the relative location of BUFFER in SOURCE.
+   Both lines and columns start numbering at one, i.e. the first
+   column/line is the column/line number 1.  */
+
+int pk_compile_expression_with_loc (pk_compiler pkc, const char *buffer,
+                                    const char *source,
+                                    uint32_t line, uint32_t column,
+                                    const char **end, pk_val *val,
+                                    pk_val *exit_exception)
+    LIBPOKE_API LIBPOKE_NONNULL (1, 2, 8);
 
 /* Load a module using the given compiler.
 
-   If the module cannot be loaded, return PK_ERROR.  Otherwise,
+   If the given module has been already be loaded, it is loaded again
+   and all the definitions are re-defined.  This is the same behavior
+   than the `load' Poke language construction.
+
+   EXIT_EXCEPTION is a pointer to a pk_val variable that is set to an
+   Exception value if the moudle load results in an unhandled exception,
+   PK_NULL otherwise.
+
+   Return PK_ERROR in case of a compilation error.  Otherwise,
    return PK_OK.  */
 
-int pk_load (pk_compiler pkc, const char *module) LIBPOKE_API;
+int pk_load (pk_compiler pkc, const char *module,
+             pk_val *exit_exception) LIBPOKE_API LIBPOKE_NONNULL (1, 2, 3);
 
 /* Print a disassembly of a named function.
 
@@ -277,6 +355,22 @@ void pk_reset_profile (pk_compiler pkc) LIBPOKE_API;
 
 void pk_set_quiet_p (pk_compiler pkc, int quiet_p) LIBPOKE_API;
 
+/* Set the DEBUG_P flag in the compiler.  If this flag is set, the
+   incremental compiler:
+
+   - Creates and saves the abstract syntax tree (AST) intermediate
+     representation of the last compiled program, expression or
+     statement, whose textual representation is then available using
+     the pk_get_debug_ast service.  */
+
+void pk_set_debug_p (pk_compiler pkc, int debug_p) LIBPOKE_API;
+
+/* Get a printable representation of the last program, expression or
+   statement compiled in debug mode.  This function returns NULL if no
+   such program has been compiled.  */
+
+const char *pk_get_debug_ast (pk_compiler pkc) LIBPOKE_API;
+
 /* Install a handler for alien tokens in the incremental compiler.
    The handler gets a string with the token identifier (for $foo it
    would get `foo') and should return a pk_alien_token struct with the
@@ -321,9 +415,19 @@ struct pk_alien_token
 
 typedef struct pk_alien_token *(*pk_alien_token_handler_fn) (const char *id,
                                                              char **errmsg);
+
+typedef struct pk_alien_token *(*pk_alien_dtoken_handler_fn) (char delim,
+                                                              const char *id,
+                                                              char **errmsg);
+
+/* Register a handler for simple $foo::bar alien tokens.  */
 void pk_set_alien_token_fn (pk_compiler pkc,
                             pk_alien_token_handler_fn cb) LIBPOKE_API;
 
+/* Register a handler for delimited $<...> alien tokens.  */
+void pk_set_alien_dtoken_fn (pk_compiler pkc,
+                             pk_alien_dtoken_handler_fn cb) LIBPOKE_API;
+
 /* Set the LEXICAL_CUCKOLDING_P flag in the compiler.  If this flag is
    set, alien tokens are recognized and processed by calling the
    handler installed by pk_set_alien_token_fn.
@@ -351,16 +455,23 @@ void pk_set_lexical_cuckolding_p (pk_compiler pkc,
 char *pk_completion_function (pk_compiler pkc,
                               const char *text, int state) LIBPOKE_API;
 
-/* Complete the tag of an IOS.
+/* Complete the name of an IOS.
+
+   PREFIX is a prefix appended to the IOS handlers as they are
+   checked.  This is to establish the context where an IOS is to be
+   recognized.  For example, for an application that uses the
+   $<HANDLER> notation for accessing IO spaces, the prefix would be
+   "$<".
 
    TEXT is the partial word to be completed.  STATE is zero the first
    time the function is called and non-zero for each subsequent call.
 
-   On each call, the function returns the tag of an IOS for which
+   On each call, the function returns the name of an IOS for which
    that tag and TEXT share a common substring. It returns NULL to
-   indicate that there are no more such tags.  */
+   indicate that there are no more such names.  */
 
 char *pk_ios_completion_function (pk_compiler pkc,
+                                  const char *prefix,
                                   const char *x, int state) LIBPOKE_API;
 
 /* Return the handler operated by the given IO space.  */
@@ -376,10 +487,17 @@ pk_ios pk_ios_cur (pk_compiler pkc) LIBPOKE_API;
 
 void pk_ios_set_cur (pk_compiler pkc, pk_ios ios) LIBPOKE_API;
 
-/* Return the IO space operating the given HANDLER.  Return NULL if no
-   such space exist in the given Poke incremental compiler.  */
+/* Return the IO space operating the given non-empty HANDLER.  Based on
+   the FLAGS search use different strategies for matching the HANDLER.
 
-pk_ios pk_ios_search (pk_compiler pkc, const char *handler) LIBPOKE_API;
+   Return NULL if no such space exist in the given Poke incremental
+   compiler.  */
+
+#define PK_IOS_SEARCH_F_EXACT   0  /* Match the whole HANDLER string.  */
+#define PK_IOS_SEARCH_F_PARTIAL 1  /* Match if HANDLER is a sub-string.  */
+
+pk_ios pk_ios_search (pk_compiler pkc, const char *handler,
+                      uint32_t flags) LIBPOKE_API;
 
 /* Return the IO space having the given ID.  Return NULL if no such
    space exist in the given Poke incremental compiler.  */
@@ -452,7 +570,32 @@ void pk_ios_map (pk_compiler pkc, pk_ios_map_fn cb, void *data) LIBPOKE_API;
    determines what subset of declarations will be processed by the
    handler.
 
-   HANDLER is a function of type pk_map_decl_fn.  */
+   HANDLER is a function of type pk_map_decl_fn.  This callback gets
+   the following arguments when invoked by pk_decl_map:
+
+     KIND is one of the PK_DECL_KIND_* values defined below.
+
+     SOURCE is a NULL-terminated string somehow denoting the source of
+     the declaration.  This may be a file name, or a descriptive
+     string like "<stdin>".
+
+     NAME is a NULL-terminated string with the name of the entity
+     that is declared.
+
+     TYPE is a NULL-terminated string with the printed representation
+     of the type of the entity that is declared.  This is NULL for
+     PKL_DECL_KIND_TYPE.
+
+     FIRST_LINE, LAST_LINE, FIRST_COLUMN and LAST_COLUMN contain the
+     location of the declaration of the entity that is declared.
+
+     VALUE is either the value of the declared entity, or PK_NULL for
+     PK_DECL_KIND_TYPE.
+
+     DATA is a user-provided pointer at pk_decl_map invocation.
+
+   DATA is a pointer that is passed to the provided callback.  This
+   can be NULL.  */
 
 #define PK_DECL_KIND_VAR 0
 #define PK_DECL_KIND_FUNC 1
@@ -505,6 +648,10 @@ int pk_defvar (pk_compiler pkc, const char *varname, pk_val val) LIBPOKE_API;
    RET is set to the value returned by the function, or to NULL if
    it is a void function.
 
+   EXIT_EXCEPTION, if not NULL, is a pointer to a pk_val variable that
+   is set to an Exception value if the function call results in an
+   unhandled exception, PK_NULL otherwise.
+
    NARG is the number of arguments that follow and that are passed in
    the function call.  Note that PK_NULL values should be added for
    "not specified" actuals for formals that have default values.
@@ -519,7 +666,8 @@ int pk_defvar (pk_compiler pkc, const char *varname, pk_val val) LIBPOKE_API;
    holding the method to be passed as the _last_ argument in the
    pk_call.  */
 
-int pk_call (pk_compiler pkc, pk_val cls, pk_val *ret,
+int pk_call (pk_compiler pkc, pk_val cls,
+             pk_val *ret, pk_val *exit_exception,
              int narg, ...) LIBPOKE_API;
 
 /* Get and set properties of the incremental compiler.  */
@@ -539,6 +687,9 @@ void pk_set_oindent (pk_compiler pkc, unsigned int oindent) LIBPOKE_API;
 int pk_omaps (pk_compiler pkc) LIBPOKE_API;
 void pk_set_omaps (pk_compiler pkc, int omaps_p) LIBPOKE_API;
 
+int pk_autoremap (pk_compiler pkc) LIBPOKE_API;
+void pk_set_autoremap (pk_compiler pkc, int autoremap_p) LIBPOKE_API;
+
 enum pk_omode
   {
     PK_PRINT_FLAT,
@@ -590,7 +741,7 @@ void pk_set_pretty_print (pk_compiler pkc, int pretty_print_p) LIBPOKE_API;
    Return PK_NULL if SIZE exceeds the maximum number of bits supported
    in Poke integers.  */
 
-pk_val pk_make_int (int64_t value, int size) LIBPOKE_API;
+pk_val pk_make_int (pk_compiler pkc, int64_t value, int size) LIBPOKE_API;
 
 /* Return the numerical value stored in the given Poke integer.
    VAL should be a Poke value of the right type.  */
@@ -612,7 +763,7 @@ int pk_int_size (pk_val val) LIBPOKE_API;
    Return NULL if SIZE exceeds the maximum number of bits supported in
    Poke integers.  */
 
-pk_val pk_make_uint (uint64_t value, int size) LIBPOKE_API;
+pk_val pk_make_uint (pk_compiler pkc, uint64_t value, int size) LIBPOKE_API;
 
 /* Return the numerical value stored in the given Poke unsigned
    integer.
@@ -633,7 +784,7 @@ int pk_uint_size (pk_val val) LIBPOKE_API;
    STR is a NULL-terminated string, a copy of which will become the
    value of the Poke string.  */
 
-pk_val pk_make_string (const char *str) LIBPOKE_API;
+pk_val pk_make_string (pk_compiler pkc, const char *str) LIBPOKE_API;
 
 /* Return a NULL-terminated string with the value of the given Poke
    string.  */
@@ -653,7 +804,8 @@ const char *pk_string_str (pk_val val) LIBPOKE_API;
    If any of the arguments is not of the right type return PK_NULL.
    Otherwise return the created offset.  */
 
-pk_val pk_make_offset (pk_val magnitude, pk_val unit) LIBPOKE_API;
+pk_val pk_make_offset (pk_compiler pkc, pk_val magnitude,
+                       pk_val unit) LIBPOKE_API;
 
 /* Return the magnitude of the given Poke offset.  */
 
@@ -675,7 +827,8 @@ pk_val pk_offset_unit (pk_val val) LIBPOKE_API;
    The fields and methods in the created struct are initialized to
    PK_NULL.  */
 
-pk_val pk_make_struct (pk_val nfields, pk_val type) LIBPOKE_API;
+pk_val pk_make_struct (pk_compiler pkc, pk_val nfields,
+                       pk_val type) LIBPOKE_API;
 
 /* Get the number of fields of a struct. */
 
@@ -683,12 +836,30 @@ pk_val pk_struct_nfields (pk_val sct) LIBPOKE_API;
 
 /* Get the value of a struct field given the name of the field.
 
+   SCT is the field value.
+
+   FNAME is a C string with the name of the field in the struct.
+
    If FNAME doesn't identify a field in the struct, PK_NULL is
    returned.  */
 
 pk_val pk_struct_ref_field_value (pk_val sct, const char *fname)
   LIBPOKE_API;
 
+/* Set the value of a struct field given the name of the field.
+
+   SCT is the struct value.
+
+   FNAME is a C string with the name of the field in the struct.
+
+   VALUE is the new value for this field.
+
+   If FNAME doesn't identify a field in the struct, struct remains
+   unchanged.  */
+
+void pk_struct_ref_set_field_value (pk_val sct, const char *fname,
+                                    pk_val value) LIBPOKE_API;
+
 /* Get the bit-offset of the field of a struct, relative to the
    beginning of the struct.
 
@@ -772,7 +943,8 @@ void pk_struct_set_field_value (pk_val sct, uint64_t idx,
    ARRAY_TYPE is the type of the array.  Note, this is an array type,
    not the type of the elements.  */
 
-pk_val pk_make_array (pk_val nelem, pk_val array_type) LIBPOKE_API;
+pk_val pk_make_array (pk_compiler pkc, pk_val nelem,
+                      pk_val array_type) LIBPOKE_API;
 
 /* Get the number of elements in the given array value, as an
    uint<64>.  */
@@ -838,7 +1010,8 @@ uint64_t pk_sizeof (pk_val val) LIBPOKE_API;
    SIGNED_P is an int<32> with a boolean specifying whether the type
    is signed or not.  */
 
-pk_val pk_make_integral_type (pk_val size, pk_val signed_p) LIBPOKE_API;
+pk_val pk_make_integral_type (pk_compiler pkc, pk_val size,
+                              pk_val signed_p) LIBPOKE_API;
 
 /* Return an uint<64> containing the size, in bits, of the given
    integral type.  */
@@ -854,13 +1027,13 @@ pk_val pk_integral_type_signed_p (pk_val type) LIBPOKE_API;
 
 /* Build and return the string type.  */
 
-pk_val pk_make_string_type (void) LIBPOKE_API;
+pk_val pk_make_string_type (pk_compiler pkc) LIBPOKE_API;
 
 /* The `any' type.  */
 
 /* Build and return the `any' type.  */
 
-pk_val pk_make_any_type (void) LIBPOKE_API;
+pk_val pk_make_any_type (pk_compiler pkc) LIBPOKE_API;
 
 /* Offset types.  */
 
@@ -870,9 +1043,13 @@ pk_val pk_make_any_type (void) LIBPOKE_API;
    of the offset.
 
    UNIT is an uint<64> with the unit of the offset type.  The unit is
-   a multiple of the base unit, which is the bit.  */
+   a multiple of the base unit, which is the bit.
 
-pk_val pk_make_offset_type (pk_val base_type, pk_val unit) LIBPOKE_API;
+   REF_TYPE is either PK_NULL or the type of the referenced type if the
+   offset is also a reference or pointer.  */
+
+pk_val pk_make_offset_type (pk_compiler pkc, pk_val base_type, pk_val unit,
+                            pk_val ref_type) LIBPOKE_API;
 
 /* Get the base type of a given offset type.  */
 
@@ -894,10 +1071,11 @@ pk_val pk_offset_type_unit (pk_val type) LIBPOKE_API;
 
    FTYPES is a C array containing the types of each struct field. */
 
-pk_val pk_make_struct_type (pk_val nfields, pk_val name, pk_val *fnames,
-                            pk_val *ftypes) LIBPOKE_API;
+pk_val pk_make_struct_type (pk_compiler pkc, pk_val nfields, pk_val name,
+                            pk_val *fnames, pk_val *ftypes) LIBPOKE_API;
 
-/* Get the type of a struct.  */
+/* Get the type of a struct.
+   This function is DEPRECATED; please use pk_typeof instead.  */
 
 pk_val pk_struct_type (pk_val sct) LIBPOKE_API;
 
@@ -907,8 +1085,8 @@ void pk_allocate_struct_attrs (pk_val nfields,
 pk_val **fnames, pk_val **ftypes) LIBPOKE_API;
 
 /* Get the name of a struct type.
-
-   If the struct type is anonymous, PK_NULL is returned.  */
+   If the struct type is anonymous, PK_NULL is returned.
+   This function is DEPRECATED; please use pk_type_name instead.  */
 
 pk_val pk_struct_type_name (pk_val type) LIBPOKE_API;
 
@@ -971,18 +1149,22 @@ void pk_struct_type_set_ftype (pk_val type, uint64_t idx,
 /* Build and return an array type.
 
    ETYPE is the type of the elements of the array.
+   BOUND is either:
+   - A ()any closure returning PK_NULL to denote an unbounded array.
+   - A closure ()uint<64> to denote an array type bounded by number of
+     elements.
+   - A closure ()offset<uint<64>,1> to denote an array type bounded by
+     size.  */
 
-   BOUND is an uint<64> with the number of elements of the
-   array... XXX or a closure?? cant remember. At the moment the PKL
-   compiler generates PVM_NULL for this... */
-
-pk_val pk_make_array_type (pk_val etype, pk_val bound) LIBPOKE_API;
+pk_val pk_make_array_type (pk_compiler pkc, pk_val etype,
+                           pk_val bound) LIBPOKE_API;
 
 /* Get the type of the elements of the given array type.  */
 
 pk_val pk_array_type_etype (pk_val type) LIBPOKE_API;
 
-/* Get the bound of the given array type.  */
+/* Get the bound of the given array type.
+   This is always PK_NULL at this point.  */
 
 pk_val pk_array_type_bound (pk_val type) LIBPOKE_API;
 
@@ -1075,10 +1257,16 @@ pk_val pk_typeof (pk_val val) LIBPOKE_API;
 #define PK_TYPE_ARRAY   5
 #define PK_TYPE_STRUCT  6
 #define PK_TYPE_CLOSURE 7
-#define PK_TYPE_ANY     8
+#define PK_TYPE_VOID    8
 
 int pk_type_code (pk_val val) LIBPOKE_API;
 
+/* Given a type value, return its name.
+
+   If the provided type is anonymous, PK_NULL is returned.  */
+
+pk_val pk_type_name (pk_val type) LIBPOKE_API;
+
 /* Given a value, return its kind.  */
 
 #define PK_VAL_UNKNOWN 0
@@ -1110,17 +1298,6 @@ int pk_val_equal_p (pk_val val1, pk_val val2) LIBPOKE_API;
 void pk_print_val (pk_compiler pkc, pk_val val,
                    pk_val *exit_exception) LIBPOKE_API;
 
-/* Print the given value using the given settings.
-   See above for an explanation for EXIT_EXCEPTION.  */
-
-#define PK_PRINT_F_MAPS   1  /* Output value and element offsets.  */
-#define PK_PRINT_F_PPRINT 2  /* Use pretty-printers.  */
-
-void pk_print_val_with_params (pk_compiler pkc, pk_val val,
-                               int depth, int mode, int base,
-                               int indent, int acutoff,
-                               uint32_t flags, pk_val *exit_exception) LIBPOKE_API;
-
 /* Foreign IO devices interface.  */
 
 typedef uint64_t pk_iod_off;
@@ -1133,6 +1310,7 @@ typedef uint64_t pk_iod_off;
 #define PK_IOD_ENOMEM -4
 #define PK_IOD_EOF    -5
 #define PK_IOD_EINVAL -6
+#define PK_IOD_EMMAP  -7
 
 struct pk_iod_if
 {
@@ -1143,51 +1321,44 @@ struct pk_iod_if
   /* Determine whether the provided HANDLER is recognized as a valid
      device spec by this backend, and if so, return its normalized
      form (caller will free).  In case of error, return NULL.  This function
-     sets the ERROR to error code or to IOD_OK.  */
-
+     sets the ERROR to error code or to PK_IOD_OK.  */
   char * (*handler_normalize) (const char *handler, uint64_t flags, int* error);
 
   /* Open a device using the provided HANDLER.  Return the opened
      device, or NULL in case of errors.  Set the ERROR to error code or to
-     IOD_OK.  Note that this function assumes that HANDLER is recognized as a
-     handler by the backend.  DATA is the `data' pointer below, intended to be
-     used as an IOD-specific payload specified by the author of IOD.  */
-
+     PK_IOD_OK.  Note that this function assumes that HANDLER is recognized as
+     a handler by the backend.  DATA is the `data' pointer below, intended to
+     be used as an IOD-specific payload specified by the author of IOD.  */
   void * (*open) (const char *handler, uint64_t flags, int *error, void *data);
 
   /* Close the given device.  Return the error code if there was an error
-     during the operation, IOD_OK otherwise.  */
-
+     during the operation, PK_IOD_OK otherwise.  */
   int (*close) (void *dev);
 
   /* Read a small byte buffer from the given device at the given byte offset.
-     Return 0 on success, or IOD_EOF on error, including on short reads.  */
-
+     Return PK_IOD_OK on success, or PK_IOD_EOF on error, including on short
+     reads.  */
   int (*pread) (void *dev, void *buf, size_t count, pk_iod_off offset);
 
   /* Write a small byte buffer to the given device at the given byte offset.
-     Return 0 on success, or IOD_EOF on error, including short writes.  */
-
+     Return PK_IOS_OK on success, or PK_IOD_EOF on error, including short
+     writes.  */
   int (*pwrite) (void *dev, const void *buf, size_t count, pk_iod_off offset);
 
   /* Return the flags of the device, as it was opened.  */
-
   uint64_t (*get_flags) (void *dev);
 
   /* Return the size of the device, in bytes.  */
-
   pk_iod_off (*size) (void *dev);
 
   /* If called on a in-stream, free the buffer before OFFSET.  If called on
      an out-stream, flush the data till OFFSET and free the buffer before
-     OFFSET.  Otherwise, do not do anything.  Return IOS_OK in success and
+     OFFSET.  Otherwise, do not do anything.  Return PK_IOS_OK in success and
      an error code on failure.  */
-
   int (*flush) (void *dev, pk_iod_off offset);
 
   /* IOD-specific data payload that will be passed to OPEN function.  If not
      used, it should be NULL.  */
-
   void *data;
 };
 
@@ -1205,4 +1376,10 @@ struct pk_iod_if
 int pk_register_iod (pk_compiler pkc, struct pk_iod_if *iod_if)
   LIBPOKE_API;
 
+/* Given a string, determine whether it corresponds to a Poke
+   keyword.  */
+
+int pk_keyword_p (pk_compiler pkc, const char *str)
+  LIBPOKE_API;
+
 #endif /* ! LIBPOKE_H */
