@@ -1,6 +1,6 @@
 /* pk-val.c - poke values.  */
 
-/* Copyright (C) 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,7 +24,7 @@
 #include "libpoke.h"
 
 pk_val
-pk_make_int (int64_t value, int size)
+pk_make_int (pk_compiler pkc __attribute__ ((unused)), int64_t value, int size)
 {
   return pvm_make_signed_integral (value, size);
 }
@@ -48,7 +48,8 @@ pk_int_size (pk_val val)
 }
 
 pk_val
-pk_make_uint (uint64_t value, int size)
+pk_make_uint (pk_compiler pkc __attribute__ ((unused)), uint64_t value,
+              int size)
 {
   return pvm_make_unsigned_integral (value, size);
 }
@@ -72,7 +73,7 @@ pk_uint_size (pk_val val)
 }
 
 pk_val
-pk_make_string (const char *str)
+pk_make_string (pk_compiler pkc __attribute__ ((unused)), const char *str)
 {
   return pvm_make_string (str);
 }
@@ -84,7 +85,8 @@ pk_string_str (pk_val val)
 }
 
 pk_val
-pk_make_offset (pk_val magnitude, pk_val unit)
+pk_make_offset (pk_compiler pkc __attribute__ ((unused)), pk_val magnitude,
+                pk_val unit)
 {
   if (!PVM_IS_INTEGRAL (magnitude)
       || !PVM_IS_ULONG (unit)
@@ -92,7 +94,11 @@ pk_make_offset (pk_val magnitude, pk_val unit)
       || PVM_VAL_ULONG_SIZE (unit) != 64)
     return PK_NULL;
   else
-    return pvm_make_offset (magnitude, unit);
+    {
+      pvm_val type = pvm_make_offset_type (pvm_typeof (magnitude),
+                                           unit, PVM_NULL /* ref_type */);
+      return pvm_make_offset (magnitude, type);
+    }
 }
 
 pk_val
@@ -104,7 +110,8 @@ pk_offset_magnitude (pk_val val)
 pk_val
 pk_offset_unit (pk_val val)
 {
-  return PVM_VAL_OFF_UNIT (val);
+  pvm_val val_type = PVM_VAL_OFF_TYPE (val);
+  return PVM_VAL_TYP_O_UNIT (val_type);
 }
 
 int
@@ -169,10 +176,10 @@ pk_val_offset (pk_val val)
   /* XXX "upunit" properly so we get a nice unit, not just bytes or
      bits.  */
   if (bit_offset % 8 == 0)
-    return pvm_make_offset (pvm_make_ulong (bit_offset / 8, 64),
-                            pvm_make_ulong (8, 64));
+    return pk_make_offset (/*pkc*/ NULL, pvm_make_ulong (bit_offset / 8, 64),
+                           pvm_make_ulong (8, 64));
   else
-    return pvm_make_offset (val_offset, pvm_make_ulong (1, 64));
+    return pk_make_offset (/*pkc*/ NULL, val_offset, pvm_make_ulong (1, 64));
 }
 
 void
@@ -180,12 +187,14 @@ pk_val_set_offset (pk_val val, pk_val off)
 {
   uint64_t boff;
 
-  if (!PVM_IS_OFF (off))
-    return;
+  if (PVM_IS_OFF (off))
+    {
+      pvm_val off_type = PVM_VAL_OFF_TYPE (off);
 
-  boff = PVM_VAL_INTEGRAL (PVM_VAL_OFF_MAGNITUDE (off))
-         * PVM_VAL_ULONG (PVM_VAL_OFF_UNIT (off));
-  PVM_VAL_SET_OFFSET (val, pvm_make_ulong (boff, 64));
+      boff = PVM_VAL_INTEGRAL (PVM_VAL_OFF_MAGNITUDE (off))
+        * PVM_VAL_ULONG (PVM_VAL_TYP_O_UNIT (off_type));
+      PVM_VAL_SET_OFFSET (val, pvm_make_ulong (boff, 64));
+    }
 }
 
 pk_val
@@ -223,13 +232,28 @@ pk_type_code (pk_val val)
       return PK_TYPE_OFFSET;
     case PVM_TYPE_CLOSURE:
       return PK_TYPE_CLOSURE;
-    case PVM_TYPE_ANY:
-      return PK_TYPE_ANY;
+    case PVM_TYPE_VOID:
+      return PK_TYPE_VOID;
     default:
       return PK_TYPE_UNKNOWN;
     }
 }
 
+pk_val
+pk_type_name (pk_val type)
+{
+  switch (PVM_VAL_TYP_CODE (type))
+    {
+    case PVM_TYPE_STRUCT:
+      return PVM_VAL_TYP_S_NAME (type);
+      break;
+    default:
+      break;
+    }
+
+  return PK_NULL;
+}
+
 int
 pk_val_kind (pk_val val)
 {
@@ -260,7 +284,8 @@ pk_val_equal_p (pk_val val1, pk_val val2)
 }
 
 pk_val
-pk_make_struct (pk_val nfields, pk_val type)
+pk_make_struct (pk_compiler pkc __attribute__ ((unused)), pk_val nfields,
+                pk_val type)
 {
   return pvm_make_struct (nfields, pvm_make_ulong (0, 64), type);
 }
@@ -277,6 +302,13 @@ pk_struct_ref_field_value (pk_val sct, const char *fname)
   return pvm_ref_struct_cstr (sct, fname);
 }
 
+void
+pk_struct_ref_set_field_value (pk_val sct, const char *fname,
+                               pk_val value)
+{
+  pvm_ref_set_struct_cstr (sct, fname, value);
+}
+
 pk_val
 pk_struct_field_boffset (pk_val sct, uint64_t idx)
 {
@@ -326,13 +358,15 @@ pk_struct_set_field_value (pk_val sct, uint64_t idx, pk_val value)
 }
 
 pk_val
-pk_make_array (pk_val nelem, pk_val array_type)
+pk_make_array (pk_compiler pkc __attribute__ ((unused)), pk_val nelem,
+               pk_val array_type)
 {
   return pvm_make_array (nelem, array_type);
 }
 
 pk_val
-pk_make_integral_type (pk_val size, pk_val signed_p)
+pk_make_integral_type (pk_compiler pkc __attribute__ ((unused)), pk_val size,
+                       pk_val signed_p)
 {
   return pvm_make_integral_type (size, signed_p);
 }
@@ -350,15 +384,16 @@ pk_integral_type_signed_p (pk_val type)
 }
 
 pk_val
-pk_make_string_type (void)
+pk_make_string_type (pk_compiler pkc __attribute__ ((unused)))
 {
   return pvm_make_string_type ();
 }
 
 pk_val
-pk_make_offset_type (pk_val base_type, pk_val unit)
+pk_make_offset_type (pk_compiler pkc __attribute__ ((unused)),
+                     pk_val base_type, pk_val unit, pk_val ref_type)
 {
-  return pvm_make_offset_type (base_type, unit);
+  return pvm_make_offset_type (base_type, unit, ref_type);
 }
 
 pk_val
@@ -374,13 +409,8 @@ pk_offset_type_unit (pk_val type)
 }
 
 pk_val
-pk_make_any_type (void)
-{
-  return pvm_make_any_type ();
-}
-
-pk_val
-pk_make_struct_type (pk_val nfields, pk_val name, pk_val *fnames, pk_val *ftypes)
+pk_make_struct_type (pk_compiler pkc __attribute__ ((unused)), pk_val nfields,
+                     pk_val name, pk_val *fnames, pk_val *ftypes)
 {
   return pvm_make_struct_type (nfields, name, fnames, ftypes);
 }
@@ -442,7 +472,8 @@ pk_struct_type_set_ftype (pk_val type, uint64_t idx, pk_val field_type)
 }
 
 pk_val
-pk_make_array_type (pk_val etype, pk_val bound)
+pk_make_array_type (pk_compiler pkc __attribute__ ((unused)), pk_val etype,
+                    pk_val bound)
 {
   return pvm_make_array_type (etype, bound);
 }
