@@ -1,6 +1,6 @@
 /* pkl-promo.c - Operand promotion phase for the poke compiler.  */
 
-/* Copyright (C) 2019, 2020, 2021 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
 
 /* This file implements a compiler phase that promotes the operands of
    expressions following language rules.  This phase expects that
-   every expression operand is anotated with its proper type.  */
+   every expression operand is annotated with its proper type.  */
 
 #include <config.h>
 
@@ -31,6 +31,21 @@
 /* Note the following macro evaluates the arguments twice!  */
 #define MAX(A,B) ((A) > (B) ? (A) : (B))
 
+/* Promote a given node A to an `any' type.  */
+
+static void
+promote_to_any (pkl_ast ast, pkl_ast_node *a, int *restart)
+{
+  pkl_ast_node any_type = pkl_ast_make_any_type (ast);
+  pkl_ast_loc loc = PKL_AST_LOC (*a);
+
+  *a = pkl_ast_make_cast (ast, any_type, ASTDEREF (*a));
+  PKL_AST_TYPE (*a) = ASTREF (any_type);
+  PKL_AST_LOC (*a) = loc;
+  *a = ASTREF (*a);
+  *restart = 1;
+}
+
 /* Promote a given node A to an integral type of width SIZE and sign
    SIGN, if possible.  Put the resulting node in A.  Return 1 if the
    promotion was successful, 0 otherwise.  */
@@ -81,6 +96,7 @@ static int
 promote_offset (pkl_ast ast,
                 size_t size, int sign,
                 pkl_ast_node unit,
+                pkl_ast_node ref_type,
                 pkl_ast_node *a,
                 int *restart)
 {
@@ -96,6 +112,7 @@ promote_offset (pkl_ast ast,
       int a_type_base_type_sign = PKL_AST_TYPE_I_SIGNED_P (a_type_base_type);
 
       int different_units = 1;
+      int different_ref_types = 1;
 
       /* If the offset units happen to be integer nodes, we can
          determine whether they are equal right away.  */
@@ -105,9 +122,19 @@ promote_offset (pkl_ast ast,
               == PKL_AST_INTEGER_VALUE (unit)))
         different_units = 0;
 
+      /* Determine whether both offset types do not have a referred
+         type, or if the referred type is the same.  */
+      if ((!PKL_AST_TYPE_O_REF_TYPE (a_type) && !ref_type)
+          || (ref_type
+              && PKL_AST_TYPE_O_REF_TYPE (a_type)
+              && !pkl_ast_type_equal_p (PKL_AST_TYPE_O_REF_TYPE (a_type),
+                                        ref_type)))
+        different_ref_types = 0;
+
       if (a_type_base_type_size != size
           || a_type_base_type_sign != sign
-          || different_units)
+          || different_units
+          || different_ref_types)
         {
           pkl_ast_loc loc = PKL_AST_LOC (*a);
           pkl_ast_node base_type
@@ -115,7 +142,7 @@ promote_offset (pkl_ast ast,
           pkl_ast_node unit_type
             = pkl_ast_make_integral_type (ast, 64, 0);
           pkl_ast_node type
-            = pkl_ast_make_offset_type (ast, base_type, unit);
+            = pkl_ast_make_offset_type (ast, base_type, unit, NULL /* ref_type */);
 
           PKL_AST_TYPE (unit) = ASTREF (unit_type);
           PKL_AST_LOC (base_type) = loc;
@@ -166,7 +193,10 @@ promote_array (pkl_ast ast,
 
   /* The case of static array types (bounded by constant number of
      elements) is handled in type equivalence.  No cast is needed.  */
-  if (bound && PKL_AST_CODE (bound) == PKL_AST_INTEGER)
+  if (bound
+      && PKL_AST_CODE (bound) == PKL_AST_INTEGER
+      && from_bound
+      && PKL_AST_CODE (from_bound) == PKL_AST_INTEGER)
     return 1;
 
   /* Unbounded array to unbounded array doesn't require of any
@@ -204,7 +234,7 @@ promote_node (pkl_ast ast,
 
   /* Note how we call promote array even if pkl_ast_type_equal_p
      declares that both types are equal.  This is because
-     pkl_ast_type_equal_p doesnt' take into account the array
+     pkl_ast_type_equal_p doesn't take into account the array
      dimensions.  */
   if (!pkl_ast_type_equal_p (node_type, type)
       || PKL_AST_TYPE_CODE (type) == PKL_TYPE_ARRAY)
@@ -223,12 +253,13 @@ promote_node (pkl_ast ast,
           {
             pkl_ast_node base_type = PKL_AST_TYPE_O_BASE_TYPE (type);
             pkl_ast_node unit = PKL_AST_TYPE_O_UNIT (type);
+            pkl_ast_node ref_type = PKL_AST_TYPE_O_REF_TYPE (type);
 
             size_t size = PKL_AST_TYPE_I_SIZE (base_type);
             int signed_p = PKL_AST_TYPE_I_SIGNED_P (base_type);
 
             if (!promote_offset (ast,
-                                 size, signed_p, unit,
+                                 size, signed_p, unit, ref_type,
                                  node,
                                  restart))
               goto error;
@@ -326,12 +357,12 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_div)
             PKL_AST_LOC (unit_bit) = PKL_AST_LOC (exp);
 
             if (!promote_offset (PKL_PASS_AST,
-                                 size, signed_p, unit_bit,
+                                 size, signed_p, unit_bit, NULL /* ref_type */,
                                  &PKL_AST_EXP_OPERAND (exp, 0), &restart1))
               goto error;
 
             if (!promote_offset (PKL_PASS_AST,
-                                 size, signed_p, unit_bit,
+                                 size, signed_p, unit_bit, NULL /* ref_type */,
                                  &PKL_AST_EXP_OPERAND (exp, 1), &restart2))
               goto error;
 
@@ -341,16 +372,22 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_div)
           }
         else
           {
-            int restart;
+            int restart1, restart2;
             pkl_ast_node op1_base_type = PKL_AST_TYPE_O_BASE_TYPE (op1_type);
 
+            if (!promote_node (PKL_PASS_AST,
+                               &PKL_AST_EXP_OPERAND (exp, 0),
+                               PKL_AST_TYPE (PKL_PASS_NODE),
+                               &restart1))
+              goto error;
+
             if (!promote_node (PKL_PASS_AST,
                                &PKL_AST_EXP_OPERAND (exp, 1),
                                op1_base_type,
-                               &restart))
+                               &restart2))
               goto error;
 
-            PKL_PASS_RESTART = restart;
+            PKL_PASS_RESTART = restart1 || restart2;
           }
         break;
       }
@@ -571,7 +608,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_mul)
       else if (PKL_AST_TYPE_CODE (op_type) == PKL_TYPE_STRING)
         ;
       else
-        assert (0);
+        PK_UNREACHABLE ();
     }
 
   PKL_PASS_DONE;
@@ -593,6 +630,7 @@ PKL_PHASE_END_HANDLER
            ARRAY    x ARRAY    -> BOOL
            STRUCT   x STRUCT   -> BOOL
            FUNCTION x FUNCTION -> BOOL
+           ANY      x ANY      -> BOOL
 
    In the I x I -> I configuration, the types of the operands are
    promoted in a way both operands end having the same type, following
@@ -611,8 +649,24 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_rela)
   pkl_ast_node op2 = PKL_AST_EXP_OPERAND (exp, 1);
   pkl_ast_node op1_type = PKL_AST_TYPE (op1);
   pkl_ast_node op2_type = PKL_AST_TYPE (op2);
+  int op1_type_code = PKL_AST_TYPE_CODE (op1_type);
+  int op2_type_code = PKL_AST_TYPE_CODE (op2_type);
 
-  if (PKL_AST_TYPE_CODE (op1_type) != PKL_AST_TYPE_CODE (op2_type))
+  if (op1_type_code == PKL_TYPE_ANY || op2_type_code == PKL_TYPE_ANY)
+    {
+      int restart1, restart2;
+
+      /* Promote the non-any operand to `any'.  */
+      if (op1_type_code != PKL_TYPE_ANY)
+        promote_to_any (PKL_PASS_AST, &PKL_AST_EXP_OPERAND (exp, 0), &restart1);
+      if (op2_type_code != PKL_TYPE_ANY)
+        promote_to_any (PKL_PASS_AST, &PKL_AST_EXP_OPERAND (exp, 1), &restart2);
+
+      PKL_PASS_RESTART = restart1 || restart2;
+      PKL_PASS_DONE;
+    }
+
+  if (op1_type_code != op2_type_code)
     goto error;
 
   /* Handle integral struct operands.  */
@@ -663,12 +717,12 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_rela)
         PKL_AST_LOC (unit_bit) = PKL_AST_LOC (exp);
 
         if (!promote_offset (PKL_PASS_AST,
-                             size, signed_p, unit_bit,
+                             size, signed_p, unit_bit, NULL /* ref_type */,
                              &PKL_AST_EXP_OPERAND (exp, 0), &restart1))
           goto error;
 
         if (!promote_offset (PKL_PASS_AST,
-                             size, signed_p, unit_bit,
+                             size, signed_p, unit_bit, NULL /* ref_type */,
                              &PKL_AST_EXP_OPERAND (exp, 1), &restart2))
           goto error;
 
@@ -682,8 +736,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_rela)
     case PKL_TYPE_ARRAY:
       /* Fallthrough.  */
     case PKL_TYPE_STRUCT:
-      /* Nothing to do.  */
-      break;
+      /* Fallthrough.  */
     case PKL_TYPE_FUNCTION:
       /* Nothing to do.  */
       break;
@@ -743,7 +796,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_bshiftpow)
         int signed_p = PKL_AST_TYPE_I_SIGNED_P (base_type);
 
         if (!promote_offset (PKL_PASS_AST,
-                             size, signed_p, unit,
+                             size, signed_p, unit, NULL /* ref_type */,
                              &PKL_AST_EXP_OPERAND (exp, 0), &restart1)
             || !promote_integral (PKL_PASS_AST, 32, 0,
                                   &PKL_AST_EXP_OPERAND (exp, 1), &restart2))
@@ -753,7 +806,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_bshiftpow)
         break;
       }
     default:
-      assert (0);
+      PK_UNREACHABLE ();
     }
 
   PKL_PASS_DONE;
@@ -833,22 +886,43 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_unary)
 }
 PKL_PHASE_END_HANDLER
 
-/* Handler for promoting indexes in indexers to unsigned 64 bit
-   values.  */
+/* Handler for promoting integral indexes in indexers:
+
+   - Integral indexes are promoted to uint<64>.
+   - Offset indexes are promoted to offset<uint<64>,1>.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_indexer)
 {
   int restart;
   pkl_ast_node node = PKL_PASS_NODE;
+  pkl_ast_node index = PKL_AST_INDEXER_INDEX (node);
+  pkl_ast_node index_type = PKL_AST_TYPE (index);
 
-  if (!promote_integral (PKL_PASS_AST, 64, 0,
-                         &PKL_AST_INDEXER_INDEX (node), &restart))
+  if (PKL_AST_TYPE_CODE (index_type) == PKL_TYPE_INTEGRAL
+      && !promote_integral (PKL_PASS_AST, 64, 0,
+                            &PKL_AST_INDEXER_INDEX (node), &restart))
     {
       PKL_ICE (PKL_AST_LOC (node),
-               "couldn't promote indexer subscript");
+               "couldn't promote integral indexer subscript");
       PKL_PASS_ERROR;
     }
 
+  if (PKL_AST_TYPE_CODE (index_type) == PKL_TYPE_OFFSET)
+    {
+      pkl_ast_node unit_bit = pkl_ast_make_integer (PKL_PASS_AST, 1);
+      unit_bit = ASTREF (unit_bit);
+
+      if (!promote_offset (PKL_PASS_AST, 64, 0, unit_bit, NULL /* ref_type */,
+                           &PKL_AST_INDEXER_INDEX (node), &restart))
+        {
+          PKL_ICE (PKL_AST_LOC (node),
+                   "couldn't promote offset indexer subscript");
+          PKL_PASS_ERROR;
+        }
+
+      pkl_ast_node_free (unit_bit);
+    }
+
   PKL_PASS_RESTART = restart;
 }
 PKL_PHASE_END_HANDLER
@@ -941,10 +1015,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_type_array)
         pkl_ast_node unit_bit = pkl_ast_make_integer (PKL_PASS_AST, 1);
 
         unit_bit = ASTREF (unit_bit);
-        PKL_AST_LOC (unit_bit) = PKL_AST_LOC (PKL_PASS_NODE);
 
         if (!promote_offset (PKL_PASS_AST,
-                             64, 0, unit_bit,
+                             64, 0, unit_bit, NULL /* ref_type */,
                              &PKL_AST_TYPE_A_BOUND (array_type), &restart))
           {
             PKL_ICE (PKL_AST_LOC (bound),
@@ -957,7 +1030,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_type_array)
         break;
       }
     default:
-      assert (0); /* This can't happen.  */
+      PK_UNREACHABLE (); /* This can't happen.  */
     }
 
   PKL_PASS_RESTART = restart;
@@ -988,7 +1061,6 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_array_initializer)
           index_type = pkl_ast_make_integral_type (PKL_PASS_AST,
                                                    64, 0);
           PKL_AST_TYPE (index) = ASTREF (index_type);
-          PKL_AST_LOC (index_type) = PKL_AST_LOC (node);
           PKL_PASS_RESTART = 1;
         }
     }
@@ -1006,6 +1078,10 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_ass_stmt)
   pkl_ast_node lvalue_type = PKL_AST_TYPE (lvalue);
   int restart = 0;
 
+  if (!exp)
+    /* Nothing to promote.  */
+    PKL_PASS_DONE;
+
   /* At this point it is assured that exp_type is promoteable to
      lvalue_type, or typify1 wouldn't have allowed this node to
      pass.  */
@@ -1187,21 +1263,21 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_return_stmt)
 }
 PKL_PHASE_END_HANDLER
 
-/* Promote arguments to printf.  */
+/* Promote arguments to format.  */
 
-PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_print_stmt)
+PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_format)
 {
-  pkl_ast_node print_stmt = PKL_PASS_NODE;
-  pkl_ast_node print_stmt_types = PKL_AST_PRINT_STMT_TYPES (print_stmt);
-  pkl_ast_node print_stmt_args = PKL_AST_PRINT_STMT_ARGS (print_stmt);
+  pkl_ast_node format = PKL_PASS_NODE;
+  pkl_ast_node format_types = PKL_AST_FORMAT_TYPES (format);
+  pkl_ast_node format_args = PKL_AST_FORMAT_ARGS (format);
 
   pkl_ast_node type, arg;
 
-  for (arg = print_stmt_args, type = print_stmt_types;
+  for (arg = format_args, type = format_types;
        arg && type;
        arg = PKL_AST_CHAIN (arg), type = PKL_AST_CHAIN (type))
     {
-      pkl_ast_node arg_exp = PKL_AST_PRINT_STMT_ARG_EXP (arg);
+      pkl_ast_node arg_exp = PKL_AST_FORMAT_ARG_EXP (arg);
 
       /* Skip arguments without associated values.  Also skip
          arguments with declared type ANY (%v) */
@@ -1211,11 +1287,11 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_print_stmt)
           int restart;
 
           if (!promote_node (PKL_PASS_AST,
-                             &PKL_AST_PRINT_STMT_ARG_EXP (arg),
+                             &PKL_AST_FORMAT_ARG_EXP (arg),
                              type, &restart))
             {
               PKL_ICE (PKL_AST_LOC (arg),
-                       "couldn't promote printf argument initializer");
+                       "couldn't promote format argument initializer");
               PKL_PASS_ERROR;
             }
 
@@ -1275,10 +1351,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_map)
   pkl_ast_node unit_bit = pkl_ast_make_integer (PKL_PASS_AST, 1);
 
   unit_bit = ASTREF (unit_bit);
-  PKL_AST_LOC (unit_bit) = PKL_AST_LOC (PKL_PASS_NODE);
 
   if (!promote_offset (PKL_PASS_AST,
-                       64, 0, unit_bit,
+                       64, 0, unit_bit, NULL /* ref_type */,
                        &PKL_AST_MAP_OFFSET (map),
                        &restart))
     {
@@ -1355,7 +1430,8 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_struct_type_field)
   pkl_ast_node elem_type = PKL_AST_STRUCT_TYPE_FIELD_TYPE (elem);
   pkl_ast_node elem_constraint = PKL_AST_STRUCT_TYPE_FIELD_CONSTRAINT (elem);
   pkl_ast_node elem_initializer = PKL_AST_STRUCT_TYPE_FIELD_INITIALIZER (elem);
-  pkl_ast_node elem_optcond = PKL_AST_STRUCT_TYPE_FIELD_OPTCOND (elem);
+  pkl_ast_node elem_optcond_pre = PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_PRE (elem);
+  pkl_ast_node elem_optcond_post = PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_POST (elem);
   pkl_ast_node elem_label = PKL_AST_STRUCT_TYPE_FIELD_LABEL (elem);
 
   if (elem_constraint)
@@ -1404,9 +1480,9 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_struct_type_field)
       PKL_PASS_RESTART = PKL_PASS_RESTART || restart;
     }
 
-  if (elem_optcond)
+  if (elem_optcond_pre)
     {
-      pkl_ast_node optcond_type = PKL_AST_TYPE (elem_optcond);
+      pkl_ast_node optcond_type = PKL_AST_TYPE (elem_optcond_pre);
       int restart = 0;
 
       /* Handle integral structure.  */
@@ -1416,17 +1492,46 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_struct_type_field)
 
       if (PKL_AST_TYPE_CODE (optcond_type) != PKL_TYPE_INTEGRAL)
         {
-          PKL_ICE (PKL_AST_LOC (elem_optcond),
-                   "non-promoteable struct field optcond at promo time");
+          PKL_ICE (PKL_AST_LOC (elem_optcond_pre),
+                   "non-promoteable struct field pre optcond at promo time");
           PKL_PASS_ERROR;
         }
 
       if (!promote_integral (PKL_PASS_AST, 32, 1,
-                             &PKL_AST_STRUCT_TYPE_FIELD_OPTCOND (elem),
+                             &PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_PRE (elem),
                              &restart))
         {
-          PKL_ICE (PKL_AST_LOC (elem_optcond),
-                   "couldn't promote struct field optcond");
+          PKL_ICE (PKL_AST_LOC (elem_optcond_pre),
+                   "couldn't promote struct field pre optcond");
+          PKL_PASS_ERROR;
+        }
+
+      PKL_PASS_RESTART = PKL_PASS_RESTART || restart;
+    }
+
+  if (elem_optcond_post)
+    {
+      pkl_ast_node optcond_type = PKL_AST_TYPE (elem_optcond_post);
+      int restart = 0;
+
+      /* Handle integral structure.  */
+      if (PKL_AST_TYPE_CODE (optcond_type) == PKL_TYPE_STRUCT
+          && PKL_AST_TYPE_S_ITYPE (optcond_type))
+        optcond_type = PKL_AST_TYPE_S_ITYPE (optcond_type);
+
+      if (PKL_AST_TYPE_CODE (optcond_type) != PKL_TYPE_INTEGRAL)
+        {
+          PKL_ICE (PKL_AST_LOC (elem_optcond_post),
+                   "non-promoteable struct field post optcond at promo time");
+          PKL_PASS_ERROR;
+        }
+
+      if (!promote_integral (PKL_PASS_AST, 32, 1,
+                             &PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_POST (elem),
+                             &restart))
+        {
+          PKL_ICE (PKL_AST_LOC (elem_optcond_post),
+                   "couldn't promote struct field post optcond");
           PKL_PASS_ERROR;
         }
 
@@ -1449,7 +1554,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_struct_type_field)
           pkl_ast_node unit_bit = pkl_ast_make_integer (PKL_PASS_AST, 1);
 
           if (!promote_offset (PKL_PASS_AST,
-                               64, 0, unit_bit,
+                               64, 0, unit_bit, NULL /* ref_type */,
                                &PKL_AST_STRUCT_TYPE_FIELD_LABEL (elem),
                                &restart))
             {
@@ -1491,6 +1596,60 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_in)
 }
 PKL_PHASE_END_HANDLER
 
+/* Promote the arguments of some attributes.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_attr)
+{
+  pkl_ast_node exp = PKL_PASS_NODE;
+  enum pkl_ast_attr attr = PKL_AST_EXP_ATTR (exp);
+  int restart = 0;
+
+  switch (attr)
+    {
+    case PKL_AST_ATTR_REF:
+      /* The argument of the attribute, if any, must be promoted to a
+         signed 32-bit integer.  */
+      if (PKL_AST_EXP_NUMOPS (exp) < 2)
+        break;
+
+      if (!promote_integral (PKL_PASS_AST,
+                             32, 1,
+                             &PKL_AST_EXP_OPERAND (exp, 1),
+                             &restart))
+        {
+          PKL_ICE (PKL_AST_LOC (exp),
+                   "couldn't promote argument of attribute expression #%" PRIu64,
+                   PKL_AST_UID (exp));
+          PKL_PASS_ERROR;
+        }
+
+      PKL_PASS_RESTART = restart;
+      break;
+    case PKL_AST_ATTR_ELEM:
+    case PKL_AST_ATTR_EOFFSET:
+    case PKL_AST_ATTR_ESIZE:
+    case PKL_AST_ATTR_ENAME:
+      /* The argument of the attribute must be promoted to an unsigned
+         64-bit integer.  */
+      if (!promote_integral (PKL_PASS_AST,
+                             64, 0,
+                             &PKL_AST_EXP_OPERAND (exp, 1),
+                             &restart))
+        {
+          PKL_ICE (PKL_AST_LOC (exp),
+                   "couldn't promote argument of attribute expression #%" PRIu64,
+                   PKL_AST_UID (exp));
+          PKL_PASS_ERROR;
+        }
+
+      PKL_PASS_RESTART = restart;
+      break;
+    default:
+      break;
+    }
+}
+PKL_PHASE_END_HANDLER
+
 PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_cons)
 {
   pkl_ast_node cons = PKL_PASS_NODE;
@@ -1574,7 +1733,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_cons)
         break;
       }
     default:
-      assert (0);
+      PK_UNREACHABLE ();
     }
 
   PKL_PASS_DONE;
@@ -1586,6 +1745,67 @@ PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_cons)
 }
 PKL_PHASE_END_HANDLER
 
+/* When an integral value is casted to an integral struct, it shall be
+   promoted to the later integral type.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_cast)
+{
+  pkl_ast_node cast = PKL_PASS_NODE;
+  pkl_ast_node exp = PKL_AST_CAST_EXP (cast);
+  pkl_ast_node from_type = PKL_AST_TYPE (exp);
+  pkl_ast_node to_type = PKL_AST_CAST_TYPE (cast);
+
+  if (PKL_AST_TYPE_CODE (to_type) == PKL_TYPE_STRUCT)
+    {
+      pkl_ast_node itype = PKL_AST_TYPE_S_ITYPE (to_type);
+
+      if (itype
+          && PKL_AST_TYPE_CODE (from_type) == PKL_TYPE_INTEGRAL)
+        {
+          int restart = 0;
+
+          if (!promote_integral (PKL_PASS_AST,
+                                 PKL_AST_TYPE_I_SIZE (itype),
+                                 PKL_AST_TYPE_I_SIGNED_P (itype),
+                                 &PKL_AST_CAST_EXP (cast),
+                                 &restart))
+            {
+              PKL_ICE (PKL_AST_LOC (exp),
+                       "couldn't promote integral to integral struct");
+              PKL_PASS_ERROR;
+            }
+
+          PKL_PASS_RESTART = restart;
+        }
+    }
+}
+PKL_PHASE_END_HANDLER
+
+/* The second operand of an APUSH operation must be promoted to the
+   type of the elements of the first operator.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_promo_ps_op_apush)
+{
+  pkl_ast_node exp = PKL_PASS_NODE;
+  pkl_ast_node op1 = PKL_AST_EXP_OPERAND (exp, 0);
+  pkl_ast_node op1_type = PKL_AST_TYPE (op1);
+  pkl_ast_node op1_elem_type = PKL_AST_TYPE_A_ETYPE (op1_type);
+  int restart = 0;
+
+  if (!promote_node (PKL_PASS_AST,
+                     &PKL_AST_EXP_OPERAND (PKL_PASS_NODE, 1),
+                     op1_elem_type,
+                     &restart))
+    {
+      PKL_ICE (PKL_AST_LOC (PKL_AST_EXP_OPERAND (PKL_PASS_NODE, 1)),
+               "couldn't promote argument to apush");
+      PKL_PASS_ERROR;
+    }
+
+  PKL_PASS_RESTART = restart;
+}
+PKL_PHASE_END_HANDLER
+
 struct pkl_phase pkl_phase_promo =
   {
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_EQ, pkl_promo_ps_op_rela),
@@ -1601,6 +1821,7 @@ struct pkl_phase pkl_phase_promo =
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_BAND, pkl_promo_ps_op_binary_intoffstrarr),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_AND, pkl_promo_ps_op_binary),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_OR, pkl_promo_ps_op_binary),
+   PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_IMPL, pkl_promo_ps_op_binary),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_NOT, pkl_promo_ps_op_unary),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_NEG, pkl_promo_ps_op_unary),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_POS, pkl_promo_ps_op_unary),
@@ -1614,6 +1835,8 @@ struct pkl_phase pkl_phase_promo =
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_DIV, pkl_promo_ps_op_div),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_CEILDIV, pkl_promo_ps_op_div),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_IN, pkl_promo_ps_op_in),
+   PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_ATTR, pkl_promo_ps_op_attr),
+   PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_APUSH, pkl_promo_ps_op_apush),
    PKL_PHASE_PS_HANDLER (PKL_AST_FUNC_ARG, pkl_promo_ps_func_arg),
    PKL_PHASE_PS_HANDLER (PKL_AST_MAP, pkl_promo_ps_map),
    PKL_PHASE_PS_HANDLER (PKL_AST_INDEXER, pkl_promo_ps_indexer),
@@ -1622,12 +1845,13 @@ struct pkl_phase pkl_phase_promo =
    PKL_PHASE_PS_HANDLER (PKL_AST_FUNCALL, pkl_promo_ps_funcall),
    PKL_PHASE_PS_HANDLER (PKL_AST_ASS_STMT, pkl_promo_ps_ass_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_RETURN_STMT, pkl_promo_ps_return_stmt),
-   PKL_PHASE_PS_HANDLER (PKL_AST_PRINT_STMT, pkl_promo_ps_print_stmt),
+   PKL_PHASE_PS_HANDLER (PKL_AST_FORMAT, pkl_promo_ps_format),
    PKL_PHASE_PS_HANDLER (PKL_AST_IF_STMT, pkl_promo_ps_if_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_LOOP_STMT, pkl_promo_ps_loop_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_promo_ps_struct_type_field),
    PKL_PHASE_PS_HANDLER (PKL_AST_COND_EXP, pkl_promo_ps_cond_exp),
    PKL_PHASE_PS_HANDLER (PKL_AST_CONS, pkl_promo_ps_cons),
+   PKL_PHASE_PS_HANDLER (PKL_AST_CAST, pkl_promo_ps_cast),
    PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_ARRAY, pkl_promo_ps_type_array),
    PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_OFFSET, pkl_promo_ps_type_offset),
   };
