@@ -1,6 +1,6 @@
 /* pkl-trans.c - Transformation phases for the poke compiler.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 #include <string.h>
 #include <xalloc.h>
 #include <stdlib.h>
+#include <errno.h>
 
 #include "pk-utils.h"
 
@@ -35,30 +36,9 @@
 #include "pkl-trans.h"
 
 /* This file implements several transformation compiler phases which,
-   generally speaking, are restartable.
-
-   `transl' makes adjustments to the compile-time lexical environment,
-            after parsing.  This phase should run on its own pass, and
-            is not restartable.
-
-   `trans1' finishes ARRAY, STRUCT and TYPE_STRUCT nodes by
-            determining its number of elements and characteristics.
-            It also finishes OFFSET nodes by replacing certain unit
-            identifiers with factors and completes/annotates other
-            structures.  It also finishes STRING nodes.
-
-   `trans2' scans the AST and annotates nodes that are literals.
-            Henceforth any other phase relying on this information
-            should be executed after trans2.
-
-   `trans3' handles nodes that can be replaced for something else at
-            compilation-time: SIZEOF for complete types.  This phase
-            is intended to be executed short before code generation.
-
-   `trans4' is executed just before the code generation pass.
-
-   See the handlers below for details.  */
+   generally speaking, are restartable.  */
 
+/* Handling of the stack of endianness.  */
 
 #define PKL_TRANS_PAYLOAD ((pkl_trans_payload) PKL_PASS_PAYLOAD)
 
@@ -81,14 +61,17 @@
     }                                             \
   while (0)
 
-static int pkl_trans_in_functions (pkl_ast_node functions[],
-                                   int next_function,
-                                   pkl_ast_node function)
+/* Handling of the stack of function contexts.  */
+
+static int
+pkl_trans_in_functions (struct pkl_trans_function_ctx functions[],
+                        int next_function,
+                        pkl_ast_node function)
 {
   int i;
   for (i = 0; i < next_function; i++)
     {
-      if (functions[i] == function)
+      if (functions[i].node == function)
         return 1;
     }
 
@@ -103,37 +86,17 @@ static int pkl_trans_in_functions (pkl_ast_node functions[],
 #define PKL_TRANS_FUNCTION                                              \
   (PKL_TRANS_PAYLOAD->next_function == 0                                \
    ? NULL                                                               \
-   : PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function - 1])
+   : &PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function - 1])
 
-#define PKL_TRANS_FUNCTION_BACK                                         \
-  (PKL_TRANS_PAYLOAD->next_function == 0                                \
-   ? 0                                                                  \
-   : PKL_TRANS_PAYLOAD->function_back[PKL_TRANS_PAYLOAD->next_function - 1])
-
-#define PKL_TRANS_INCR_FUNCTION_BACK                                    \
-  do                                                                    \
-  {                                                                     \
-    if (PKL_TRANS_PAYLOAD->next_function > 0)                           \
-      PKL_TRANS_PAYLOAD->function_back[PKL_TRANS_PAYLOAD->next_function - 1]++; \
-  }                                                                     \
-  while (0)
-
-#define PKL_TRANS_DECR_FUNCTION_BACK                                    \
-  do                                                                    \
-  {                                                                     \
-    if (PKL_TRANS_PAYLOAD->next_function > 0)                           \
-      PKL_TRANS_PAYLOAD->function_back[PKL_TRANS_PAYLOAD->next_function - 1]--; \
-  }                                                                     \
-  while (0)
-
-#define PKL_TRANS_PUSH_FUNCTION(function)                               \
+#define PKL_TRANS_PUSH_FUNCTION(func)                                   \
   do                                                                    \
     {                                                                   \
       assert (PKL_TRANS_PAYLOAD->next_function < PKL_TRANS_MAX_FUNCTION_NEST); \
-      PKL_TRANS_PAYLOAD->function_back[PKL_TRANS_PAYLOAD->next_function] \
-        = 0;                                                            \
-      PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function++]  \
-        = (function);                                                   \
+      PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function].back = 0; \
+      PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function].ndrops = 0; \
+      PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function].npopes = 0; \
+      PKL_TRANS_PAYLOAD->functions[PKL_TRANS_PAYLOAD->next_function++].node \
+        = (func);                                                       \
     }                                                                   \
   while (0)
 
@@ -145,6 +108,36 @@ static int pkl_trans_in_functions (pkl_ast_node functions[],
     }                                                           \
   while (0)
 
+/* Handling of the stack of escapable constructions.  */
+
+#define PKL_TRANS_ESCAPABLE                                     \
+  (PKL_TRANS_PAYLOAD->next_escapable == 0                       \
+   ? NULL                                                       \
+   : &PKL_TRANS_PAYLOAD->escapables[PKL_TRANS_PAYLOAD->next_escapable - 1])
+
+#define PKL_TRANS_PUSH_ESCAPABLE(b)                             \
+  do                                                            \
+    {                                                           \
+      assert (PKL_TRANS_PAYLOAD->next_escapable < PKL_TRANS_MAX_COMP_STMT_NEST);\
+      PKL_TRANS_PAYLOAD->escapables[PKL_TRANS_PAYLOAD->next_escapable].nframes  \
+        = 0;                                                    \
+      PKL_TRANS_PAYLOAD->escapables[PKL_TRANS_PAYLOAD->next_escapable].npopes \
+        = 0;                                                    \
+      PKL_TRANS_PAYLOAD->escapables[PKL_TRANS_PAYLOAD->next_escapable++].node \
+        = (b);                                                  \
+    }                                                           \
+  while (0)
+
+#define PKL_TRANS_PUSH_UNESCAPABLE PKL_TRANS_PUSH_ESCAPABLE (NULL)
+
+#define PKL_TRANS_POP_ESCAPABLE                                 \
+  do                                                            \
+    {                                                           \
+      assert (PKL_TRANS_PAYLOAD->next_escapable > 0);           \
+      PKL_TRANS_PAYLOAD->next_escapable--;                      \
+    }                                                           \
+  while (0)
+
 /* The following handler is used in all trans phases and initializes
    the phase payload.  */
 
@@ -185,6 +178,28 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_struct)
 }
 PKL_PHASE_END_HANDLER
 
+/* Array types conform an unescapable context.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_type_array)
+{
+  PKL_TRANS_PUSH_UNESCAPABLE;
+}
+PKL_PHASE_END_HANDLER
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_type_array)
+{
+  PKL_TRANS_POP_ESCAPABLE;
+}
+PKL_PHASE_END_HANDLER
+
+/* Struct types start an unescapable context.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_type_struct)
+{
+  PKL_TRANS_PUSH_UNESCAPABLE;
+}
+PKL_PHASE_END_HANDLER
+
 /* Compute and set the number of elements, fields and declarations in
    a struct TYPE node.  */
 
@@ -192,21 +207,29 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_type_struct)
 {
   pkl_ast_node struct_type = PKL_PASS_NODE;
   pkl_ast_node t;
-  size_t nelem = 0, nfield = 0, ndecl = 0;
+  size_t nelem = 0, nfield = 0, ncfield = 0, ndecl = 0;
 
   for (t = PKL_AST_TYPE_S_ELEMS (struct_type); t;
        t = PKL_AST_CHAIN (t))
     {
       nelem++;
       if (PKL_AST_CODE (t) == PKL_AST_STRUCT_TYPE_FIELD)
-        nfield++;
+        {
+          if (PKL_AST_STRUCT_TYPE_FIELD_COMPUTED_P (t))
+            ncfield++;
+          else
+            nfield++;
+        }
       else
         ndecl++;
     }
 
   PKL_AST_TYPE_S_NELEM (struct_type) = nelem;
   PKL_AST_TYPE_S_NFIELD (struct_type) = nfield;
+  PKL_AST_TYPE_S_NCFIELD (struct_type) = ncfield;
   PKL_AST_TYPE_S_NDECL (struct_type) = ndecl;
+
+  PKL_TRANS_POP_ESCAPABLE;
 }
 PKL_PHASE_END_HANDLER
 
@@ -246,6 +269,14 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_funcall)
 }
 PKL_PHASE_END_HANDLER
 
+/* Annotate whether declaration nodes are in the body of a struct
+   type, as this implies treating them especially somewhere else.
+
+   In declaration of functions, annotate the name of the declaration
+   in the function as the name it was declared with.  This has to be
+   done in pre-order because other phases for functions require the
+   name.  */
+
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_decl)
 {
   pkl_ast_node decl = PKL_PASS_NODE;
@@ -258,23 +289,15 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_decl)
 
   if (PKL_AST_DECL_KIND (decl) == PKL_AST_DECL_KIND_FUNC)
     {
-      pkl_ast_node func = PKL_AST_DECL_INITIAL (decl);
+      pkl_ast_node name = PKL_AST_DECL_NAME (decl);
+      pkl_ast_node function = PKL_AST_DECL_INITIAL (decl);
 
-      /* Add this function to the pass stack of functions.  */
-      PKL_TRANS_PUSH_FUNCTION (func);
+      PKL_AST_FUNC_NAME (function)
+        = xstrdup (PKL_AST_IDENTIFIER_POINTER (name));
     }
 }
 PKL_PHASE_END_HANDLER
 
-PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_decl)
-{
-  pkl_ast_node decl = PKL_PASS_NODE;
-
-  if (PKL_AST_DECL_KIND (decl) == PKL_AST_DECL_KIND_FUNC)
-    PKL_TRANS_POP_FUNCTION;
-}
-PKL_PHASE_END_HANDLER
-
 /* Variables that refer to the current function (recursive calls)
    should be marked as such, so `pkl_ast_node_free' knows to not free
    the declaration.  This is to avoid loops in the AST reference
@@ -296,8 +319,11 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_var)
   pkl_ast_node var = PKL_PASS_NODE;
   pkl_ast_node decl = PKL_AST_VAR_DECL (var);
 
-  PKL_AST_VAR_FUNCTION (var) = PKL_TRANS_FUNCTION;
-  PKL_AST_VAR_FUNCTION_BACK (var) = PKL_TRANS_FUNCTION_BACK;
+  if (PKL_TRANS_FUNCTION)
+    {
+      PKL_AST_VAR_FUNCTION (var) = PKL_TRANS_FUNCTION->node;
+      PKL_AST_VAR_FUNCTION_BACK (var) = PKL_TRANS_FUNCTION->back;
+    }
 
   if (PKL_AST_DECL_KIND (decl) == PKL_AST_DECL_KIND_FUNC)
     PKL_AST_VAR_IS_RECURSIVE (var)
@@ -319,7 +345,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_var)
                                                        NULL /* args */);
 
           PKL_AST_LOC (funcall) = PKL_AST_LOC (var);
-          PKL_PASS_NODE = funcall;
+          PKL_PASS_NODE = ASTREF (funcall);
           PKL_PASS_RESTART = 1;
         }
     }
@@ -473,7 +499,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_string)
             }
           break;
         default:
-          assert (0);
+          PK_UNREACHABLE ();
         }
     }
   new_string_pointer[i] = '\0';
@@ -486,9 +512,13 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_string)
 }
 PKL_PHASE_END_HANDLER
 
-/* Determine the attribute code of attribute expressions, emitting an
-   error if the given attribute name is not defined.  Finally, turn
-   the binary expression into an unary expression.  */
+/* At this point the second operand of attribute operations is an
+   identifier with the name of the attribute (.i.e. "length" for
+   foo'length).  Determine the corresponding attribute code, emitting
+   an error if the given attribute name is not defined.  Finally, turn
+   the binary or ternary expression into an either unary or binary
+   expression, depending whether the attribute gets an argument or it
+   doesn't.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_op_attr)
 {
@@ -519,14 +549,35 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_op_attr)
 
   PKL_AST_EXP_ATTR (exp) = attr;
 
-  /* Turn the binary expression into an unary expression.  */
-  PKL_AST_EXP_NUMOPS (exp) = 1;
-  pkl_ast_node_free (PKL_AST_EXP_OPERAND (exp, 1));
+  /* Turn the expression into an unary or binary expression.  */
+  if (PKL_AST_EXP_NUMOPS (exp) == 2)
+    {
+      PKL_AST_EXP_NUMOPS (exp) = 1;
+      pkl_ast_node_free (PKL_AST_EXP_OPERAND (exp, 1));
+    }
+  else
+    {
+      PKL_AST_EXP_NUMOPS (exp) = 2;
+      pkl_ast_node_free (PKL_AST_EXP_OPERAND (exp, 1));
+      PKL_AST_EXP_OPERAND (exp, 1) = PKL_AST_EXP_OPERAND (exp, 2);
+    }
 }
 PKL_PHASE_END_HANDLER
 
-/* Set the function's first optional argument and count the number of
-   formal arguments.  */
+/* Push the function in the stack of function contexts.  Function
+   bodies also start an unescapable context.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_func)
+{
+  PKL_TRANS_PUSH_FUNCTION (PKL_PASS_NODE);
+  PKL_TRANS_PUSH_UNESCAPABLE;
+}
+PKL_PHASE_END_HANDLER
+
+/* Annotate the function's first optional argument as such, and count
+   the number of formal arguments the function gets.
+
+   Also, pop the function from the stack of function contexts.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_func)
 {
@@ -550,11 +601,23 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_func)
           break;
         }
     }
+
+  /* Remove this function from stack of functions.  */
+  PKL_TRANS_POP_FUNCTION;
+
+  /* Pop the unescapable context.  */
+  PKL_TRANS_POP_ESCAPABLE;
 }
 PKL_PHASE_END_HANDLER
 
-/* Function types from function type literals don't have the number of
-   elements set.  Do it here.  */
+/* Determine the number of formal arguments taken by functions of this
+   type and annotate it in the function type.
+
+   Determine the first optional formal argument and annotate it in the
+   function type.
+
+   Determine whether the function gets a vararg and annotate it in the
+   function type.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_type_function)
 {
@@ -655,7 +718,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_format)
   pkl_ast_node format_fmt = PKL_AST_FORMAT_FMT (format);
   char *fmt, *p;
   pkl_ast_node t, arg;
-  int ntag, nargs = 0;
+  int ntag, nargs = PKL_AST_FORMAT_NARGS (format);
   pkl_ast_node types = NULL, prev_arg = NULL;
   const char *msg = NULL;
   /* XXX this hard limit should go away.  */
@@ -849,6 +912,110 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_format)
             ntag++;
             break;
           }
+        case 'f':
+        case 'e':
+        case 'g':
+          {
+            unsigned int bits;
+
+            PKL_AST_FORMAT_ARG_FLOATING_POINT_P (arg) = 1;
+            PKL_AST_FORMAT_ARG_FLOATING_POINT_STYLE (arg) = p[1];
+
+            if (p[2] >= '0' && p[2] <= '9')
+              {
+                int next_idx;
+
+                if (p[3] >= '0' && p[3] <= '9')
+                  {
+                    bits = (p[2] - '0') * 10 + (p[3] - '0');
+                    next_idx = 4;
+                  }
+                else
+                  {
+                    bits = p[2] - '0';
+                    next_idx = 3;
+                  }
+
+                if (bits == 0 || !(/*bits == 16 ||*/ bits == 32 || bits == 64))
+                  {
+                    msg = _("invalid bit-width for a floating-point number");
+                    goto invalid_tag;
+                  }
+                PKL_AST_FORMAT_ARG_FLOATING_POINT_WIDTH (arg)
+                    = bits == 32
+                          ? PKL_AST_FORMAT_ARG_FLOATING_POINT_WIDTH_SINGLE
+                          : PKL_AST_FORMAT_ARG_FLOATING_POINT_WIDTH_DOUBLE;
+
+                if (p[next_idx] == '.')
+                  {
+                    int prec_begin_idx;
+                    char tmp;
+                    long prec;
+
+                    prec_begin_idx = ++next_idx;
+                    while (p[next_idx] >= '0' && p[next_idx] <= '9')
+                      ++next_idx;
+                    if (next_idx == prec_begin_idx)
+                      {
+                        msg = _("expected a precision number after dot");
+                        goto invalid_tag;
+                      }
+                    tmp = p[next_idx];
+                    p[next_idx] = '\0';
+                    errno = 0;
+                    prec = strtol (&p[prec_begin_idx], NULL, 10);
+                    if (errno != 0)
+                      {
+                        msg = _("invalid precision");
+                        goto invalid_tag;
+                      }
+                    PKL_AST_FORMAT_ARG_FLOATING_POINT_PREC (arg)
+                        = (unsigned)prec;
+                    p[next_idx] = tmp;
+                  }
+                else
+                  {
+                    PKL_AST_FORMAT_ARG_FLOATING_POINT_PREC (arg)
+                        = bits == 32 ? 7 : 15;
+                  }
+
+                switch (p[next_idx])
+                  {
+                  case 'b': PKL_AST_FORMAT_ARG_BASE (arg) = 2; break;
+                  case 'o': PKL_AST_FORMAT_ARG_BASE (arg) = 8; break;
+                  case 'd': PKL_AST_FORMAT_ARG_BASE (arg) = 10; break;
+                  case 'x': PKL_AST_FORMAT_ARG_BASE (arg) = 16; break;
+                  default:
+                    msg = _("invalid base");
+                    goto invalid_tag;
+                  }
+
+                if (PKL_AST_FORMAT_ARG_BASE (arg) != 10)
+                  {
+                    msg = _("only base 10 is supported for floating-point numbers");
+                    goto invalid_tag;
+                  }
+
+                atype = pkl_ast_make_integral_type (PKL_PASS_AST,
+                                                    bits, /*signed_p*/ 0);
+                types = pkl_ast_chainon (types, atype);
+
+                p += ++next_idx;
+              }
+            else
+              {
+                if (p[1] == 'f')
+                  msg = _("expected decimal digit after %f");
+                else if (p[1] == 'e')
+                  msg = _("expected decimal digit after %e");
+                else
+                  msg = _("expected decimal digit after %g");
+
+                goto invalid_tag;
+              }
+            ntag++;
+            break;
+          }
         case '<':
           /* Fallthrough.  */
         case '>':
@@ -1096,24 +1263,53 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_array)
 }
 PKL_PHASE_END_HANDLER
 
-/* Compound statements introduce a lexical level.  Update the function
-   back.  */
+/* Compound statements introduce a lexical level.  Update the current
+   function context accordingly.
+
+   If the compound statement is the first operand of an excond
+   operator, then it increases the number of exception handlers that
+   need to be eventually popped.  Update the current function context
+   accordingly */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_comp_stmt)
 {
-  PKL_TRANS_INCR_FUNCTION_BACK;
+  if (PKL_TRANS_FUNCTION)
+    PKL_TRANS_FUNCTION->back++;
+
+  if (PKL_TRANS_ESCAPABLE)
+    PKL_TRANS_ESCAPABLE->nframes++;
+
+  if (PKL_PASS_PARENT
+      && PKL_AST_CODE (PKL_PASS_PARENT) == PKL_AST_EXP
+      && PKL_AST_EXP_CODE (PKL_PASS_PARENT) == PKL_AST_OP_EXCOND)
+    {
+      if (PKL_TRANS_FUNCTION)
+        PKL_TRANS_FUNCTION->npopes++;
+
+      if (PKL_TRANS_ESCAPABLE)
+        PKL_TRANS_ESCAPABLE->npopes++;
+    }
 }
 PKL_PHASE_END_HANDLER
 
 /* FOR-IN statements introduce a lexical level if they use an iterator
-   or have a head of declarations.  Update the function back accordingly.  */
+   or have a head of declarations.  Update the current function
+   context accordingly.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_loop_stmt_iterator)
 {
-  PKL_TRANS_INCR_FUNCTION_BACK;
+  if (PKL_TRANS_FUNCTION)
+    PKL_TRANS_FUNCTION->back++;
 }
 PKL_PHASE_END_HANDLER
 
+/* FOR statements introduce a lexical level if they use an iterator
+   or have a head of declarations.  Update the current function
+   context accordingly.
+
+   FOR statements are escapable constructs, push it to current
+   escapables stack.  */
+
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_loop_stmt)
 {
   pkl_ast_node stmt = PKL_PASS_NODE;
@@ -1121,24 +1317,45 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_loop_stmt)
   if (PKL_AST_LOOP_STMT_HEAD (stmt))
     {
       assert (!PKL_AST_LOOP_STMT_ITERATOR (stmt));
-      PKL_TRANS_INCR_FUNCTION_BACK;
+      if (PKL_TRANS_FUNCTION)
+        PKL_TRANS_FUNCTION->back++;
     }
+
+  if (PKL_TRANS_FUNCTION && PKL_AST_LOOP_STMT_ITERATOR (stmt))
+    PKL_TRANS_FUNCTION->ndrops += 3;
+
+  PKL_TRANS_PUSH_ESCAPABLE (stmt);
 }
 PKL_PHASE_END_HANDLER
 
+/* Revert what we did in `pkl_trans1_pr_loop_stmt'.  */
+
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_loop_stmt)
 {
   pkl_ast_node stmt = PKL_PASS_NODE;
 
-  if (PKL_AST_LOOP_STMT_ITERATOR (stmt) || PKL_AST_LOOP_STMT_HEAD (stmt))
-    PKL_TRANS_DECR_FUNCTION_BACK;
+  if (PKL_TRANS_FUNCTION
+      && (PKL_AST_LOOP_STMT_ITERATOR (stmt) || PKL_AST_LOOP_STMT_HEAD (stmt)))
+    PKL_TRANS_FUNCTION->back--;
+
+  if (PKL_TRANS_FUNCTION
+      && (PKL_TRANS_FUNCTION && PKL_AST_LOOP_STMT_ITERATOR (stmt)))
+    PKL_TRANS_FUNCTION->ndrops -= 3;
+
+  PKL_TRANS_POP_ESCAPABLE;
 }
 PKL_PHASE_END_HANDLER
 
-/* Annotate compount statement nodes with the number of variable and
+/* Annotate compound statement nodes with the number of variable and
    function declarations occurring in the statement.
 
-   Update the function back.  */
+   Compound statements introduce a lexical level.  Update the current
+   function context and current escapable context accordingly.
+
+   If the compound statement is the first operand of an excond
+   operator, then it increases the number of exception handlers that
+   need to be eventually popped.  Update the current function context
+   accordingly */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_comp_stmt)
 {
@@ -1156,11 +1373,27 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_comp_stmt)
     }
 
   PKL_AST_COMP_STMT_NUMVARS (comp_stmt) = numvars;
-  PKL_TRANS_DECR_FUNCTION_BACK;
+
+  if (PKL_TRANS_FUNCTION)
+    PKL_TRANS_FUNCTION->back--;
+
+  if (PKL_TRANS_ESCAPABLE)
+    PKL_TRANS_ESCAPABLE->nframes--;
+
+  if (PKL_PASS_PARENT
+      && PKL_AST_CODE (PKL_PASS_PARENT) == PKL_AST_EXP
+      && PKL_AST_EXP_CODE (PKL_PASS_PARENT) == PKL_AST_OP_EXCOND)
+    {
+      if (PKL_TRANS_FUNCTION)
+        PKL_TRANS_FUNCTION->npopes--;
+
+      if (PKL_TRANS_ESCAPABLE)
+        PKL_TRANS_ESCAPABLE->npopes--;
+    }
 }
 PKL_PHASE_END_HANDLER
 
-/* Push current endianness annotation.  */
+/* Push current endianness annotation if necessary.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_struct_type_field)
 {
@@ -1172,7 +1405,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_struct_type_field)
 }
 PKL_PHASE_END_HANDLER
 
-/* Annotate struct type fields with the current endiannes annotation
+/* Annotate struct type fields with the current endianness annotation
    if necessary.  And also pop endianness.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_struct_type_field)
@@ -1188,6 +1421,119 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_struct_type_field)
 }
 PKL_PHASE_END_HANDLER
 
+/* Annotate return statements with their containing functions.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_return_stmt)
+{
+  pkl_ast_node stmt = PKL_PASS_NODE;
+
+  if (PKL_TRANS_FUNCTION)
+    {
+      PKL_AST_RETURN_STMT_FUNCTION (stmt)
+        = PKL_TRANS_FUNCTION->node; /* Note no ASTREF.  */
+      PKL_AST_RETURN_STMT_NDROPS (stmt) = PKL_TRANS_FUNCTION->ndrops;
+      PKL_AST_RETURN_STMT_NPOPES (stmt) = PKL_TRANS_FUNCTION->npopes;
+    }
+}
+PKL_PHASE_END_HANDLER
+
+/* The body of a try statement increases the number of exception
+   handlers that need to be eventually popped.  Update the current
+   function context accordingly.
+
+   try-until statements are escapable constructs.  Update escapables
+   stack accordingly.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_try_stmt_body)
+{
+  if (PKL_TRANS_FUNCTION)
+    PKL_TRANS_FUNCTION->npopes++;
+
+  if (PKL_AST_TRY_STMT_KIND (PKL_PASS_PARENT) == PKL_AST_TRY_STMT_KIND_UNTIL)
+    PKL_TRANS_PUSH_ESCAPABLE (PKL_PASS_PARENT);
+  else if (PKL_TRANS_ESCAPABLE)
+    PKL_TRANS_ESCAPABLE->npopes++;
+}
+PKL_PHASE_END_HANDLER
+
+/* The body of a try statement increases the number of exception
+   handlers that need to be eventually popped.  Update the current
+   function context accordingly.
+
+   try-until statements are escapable constructs.  Update escapables
+   stack accordingly.  try-catch statements increase the number of
+   exception handlers that need to be eventually popped.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_try_stmt_body)
+{
+  if (PKL_TRANS_FUNCTION)
+    PKL_TRANS_FUNCTION->npopes--;
+
+  if (PKL_AST_TRY_STMT_KIND (PKL_PASS_PARENT) == PKL_AST_TRY_STMT_KIND_UNTIL)
+    PKL_TRANS_POP_ESCAPABLE;
+  else if (PKL_TRANS_ESCAPABLE)
+    PKL_TRANS_ESCAPABLE->npopes--;
+}
+PKL_PHASE_END_HANDLER
+
+/* The handler of a try-catch statement introduces a new lexical
+   level for escapable constructs.  Update the current escapable
+   context.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_try_stmt_handler)
+{
+  if (PKL_AST_TRY_STMT_ARG (PKL_PASS_PARENT) && PKL_TRANS_ESCAPABLE)
+    PKL_TRANS_ESCAPABLE->nframes++;
+}
+PKL_PHASE_END_HANDLER
+
+/* Likewise.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_try_stmt_handler)
+{
+  if (PKL_AST_TRY_STMT_ARG (PKL_PASS_PARENT) && PKL_TRANS_ESCAPABLE)
+    PKL_TRANS_ESCAPABLE->nframes--;
+}
+PKL_PHASE_END_HANDLER
+
+/* try-until statements are escapable constructs.  Update escapables
+   stack accordingly.  try-catch statements increase the number of
+   exception handlers that need to be eventually dropped.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_pr_try_stmt)
+{
+  if (PKL_AST_TRY_STMT_KIND (PKL_PASS_NODE) == PKL_AST_TRY_STMT_KIND_UNTIL)
+    PKL_TRANS_PUSH_ESCAPABLE (PKL_PASS_NODE);
+}
+PKL_PHASE_END_HANDLER
+
+/* Likewise.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_try_stmt)
+{
+  if (PKL_AST_TRY_STMT_KIND (PKL_PASS_NODE) == PKL_AST_TRY_STMT_KIND_UNTIL)
+    PKL_TRANS_POP_ESCAPABLE;
+}
+PKL_PHASE_END_HANDLER
+
+/* Annotate break and continue statements with enclosing escapable
+   construct, and with their lexical nesting level with respect to the
+   enclosing escapable construct.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_break_continue_stmt)
+{
+  if (PKL_TRANS_ESCAPABLE)
+    {
+      PKL_AST_BREAK_CONTINUE_STMT_ENTITY (PKL_PASS_NODE)
+        = PKL_TRANS_ESCAPABLE->node;
+      PKL_AST_BREAK_CONTINUE_STMT_NFRAMES (PKL_PASS_NODE)
+        = PKL_TRANS_ESCAPABLE->nframes;
+      PKL_AST_BREAK_CONTINUE_STMT_NPOPES (PKL_PASS_NODE)
+        = PKL_TRANS_ESCAPABLE->npopes;
+    }
+}
+PKL_PHASE_END_HANDLER
+
 /* Annotate indexers that are themselves immediately indexed by
    another [] operator.  This is used for certain optimizations in
    `gen'.  */
@@ -1202,6 +1548,44 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_indexer)
 }
 PKL_PHASE_END_HANDLER
 
+/* Reverse list of outputs and transform them into assignments,
+   checking that they are proper l-values.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans1_ps_asm_stmt)
+{
+  pkl_ast_node asm_stmt = PKL_PASS_NODE;
+  pkl_ast_node assignments = NULL, output;
+
+  for (output = PKL_AST_ASM_STMT_OUTPUTS (asm_stmt);
+       output;
+       output = PKL_AST_CHAIN (output))
+    {
+      pkl_ast_node ass_stmt;
+
+      if (!pkl_ast_lvalue_p (output))
+        {
+          PKL_ERROR (PKL_AST_LOC (output),
+                     "asm statement output should be a l-value");
+          PKL_TRANS_PAYLOAD->errors++;
+          PKL_PASS_ERROR;
+        }
+
+      ass_stmt = pkl_ast_make_ass_stmt (PKL_PASS_AST,
+                                        output,
+                                        NULL /* exp */);
+      PKL_AST_LOC (ass_stmt) = PKL_AST_LOC (asm_stmt);
+      /* Note the reverse order.  */
+      assignments = pkl_ast_chainon (ass_stmt, assignments);
+    }
+
+  if (assignments)
+    {
+      PKL_AST_ASM_STMT_OUTPUTS (asm_stmt) = ASTREF (assignments);
+      PKL_PASS_RESTART = 1;
+    }
+}
+PKL_PHASE_END_HANDLER
+
 struct pkl_phase pkl_phase_trans1 =
   {
    PKL_PHASE_PS_HANDLER (PKL_AST_SRC, pkl_trans_ps_src),
@@ -1211,21 +1595,33 @@ struct pkl_phase pkl_phase_trans1 =
    PKL_PHASE_PS_HANDLER (PKL_AST_FUNCALL, pkl_trans1_ps_funcall),
    PKL_PHASE_PS_HANDLER (PKL_AST_STRING, pkl_trans1_ps_string),
    PKL_PHASE_PS_HANDLER (PKL_AST_VAR, pkl_trans1_ps_var),
+   PKL_PHASE_PR_HANDLER (PKL_AST_FUNC, pkl_trans1_pr_func),
    PKL_PHASE_PS_HANDLER (PKL_AST_FUNC, pkl_trans1_ps_func),
    PKL_PHASE_PS_HANDLER (PKL_AST_TRIMMER, pkl_trans1_ps_trimmer),
    PKL_PHASE_PS_HANDLER (PKL_AST_FORMAT, pkl_trans1_ps_format),
    PKL_PHASE_PR_HANDLER (PKL_AST_DECL, pkl_trans1_pr_decl),
-   PKL_PHASE_PS_HANDLER (PKL_AST_DECL, pkl_trans1_ps_decl),
    PKL_PHASE_PS_HANDLER (PKL_AST_ARRAY, pkl_trans1_ps_array),
    PKL_PHASE_PR_HANDLER (PKL_AST_COMP_STMT, pkl_trans1_pr_comp_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_COMP_STMT, pkl_trans1_ps_comp_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_LOOP_STMT_ITERATOR, pkl_trans1_ps_loop_stmt_iterator),
    PKL_PHASE_PR_HANDLER (PKL_AST_LOOP_STMT, pkl_trans1_pr_loop_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_LOOP_STMT, pkl_trans1_ps_loop_stmt),
+   PKL_PHASE_PS_HANDLER (PKL_AST_BREAK_CONTINUE_STMT, pkl_trans1_ps_break_continue_stmt),
+   PKL_PHASE_PR_HANDLER (PKL_AST_TRY_STMT, pkl_trans1_pr_try_stmt),
+   PKL_PHASE_PS_HANDLER (PKL_AST_TRY_STMT, pkl_trans1_ps_try_stmt),
+   PKL_PHASE_PR_HANDLER (PKL_AST_TRY_STMT_BODY, pkl_trans1_pr_try_stmt_body),
+   PKL_PHASE_PS_HANDLER (PKL_AST_TRY_STMT_BODY, pkl_trans1_ps_try_stmt_body),
+   PKL_PHASE_PR_HANDLER (PKL_AST_TRY_STMT_HANDLER, pkl_trans1_pr_try_stmt_handler),
+   PKL_PHASE_PS_HANDLER (PKL_AST_TRY_STMT_HANDLER, pkl_trans1_ps_try_stmt_handler),
    PKL_PHASE_PR_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_trans1_pr_struct_type_field),
    PKL_PHASE_PS_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_trans1_ps_struct_type_field),
+   PKL_PHASE_PS_HANDLER (PKL_AST_RETURN_STMT, pkl_trans1_ps_return_stmt),
    PKL_PHASE_PS_HANDLER (PKL_AST_INDEXER, pkl_trans1_ps_indexer),
+   PKL_PHASE_PS_HANDLER (PKL_AST_ASM_STMT, pkl_trans1_ps_asm_stmt),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_ATTR, pkl_trans1_ps_op_attr),
+   PKL_PHASE_PR_TYPE_HANDLER (PKL_TYPE_ARRAY, pkl_trans1_pr_type_array),
+   PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_ARRAY, pkl_trans1_ps_type_array),
+   PKL_PHASE_PR_TYPE_HANDLER (PKL_TYPE_STRUCT, pkl_trans1_pr_type_struct),
    PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_STRUCT, pkl_trans1_ps_type_struct),
    PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_FUNCTION, pkl_trans1_ps_type_function),
   };
@@ -1252,7 +1648,12 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_exp)
         break;
     }
 
-  PKL_AST_LITERAL_P (exp) = literal_p;
+  /* XXX note that POW is currently not constant-folded.  See comment
+     in pkl-fold.c */
+  if (PKL_AST_EXP_CODE (exp) == PKL_AST_OP_POW)
+    PKL_AST_LITERAL_P (exp) = 0;
+  else
+    PKL_AST_LITERAL_P (exp) = literal_p;
 }
 PKL_PHASE_END_HANDLER
 
@@ -1335,8 +1736,7 @@ PKL_PHASE_END_HANDLER
    Also, struct references that refer to parameterless methods are
    transformed into funcalls to these methods, but only if the struct
    references are not part of funcall themselves.  This is similar to
-   what is done with variable nodes. */
-
+   what is done with variable nodes.  */
 
 PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_struct_ref)
 {
@@ -1365,7 +1765,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_struct_ref)
           PKL_AST_TYPE (funcall) = ASTREF (function_rtype);
           PKL_AST_LOC (funcall) = PKL_AST_LOC (struct_ref);
 
-          PKL_PASS_NODE = funcall;
+          PKL_PASS_NODE = ASTREF (funcall);
           PKL_PASS_RESTART = 1;
         }
     }
@@ -1446,6 +1846,7 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_incrdecr)
       PKL_AST_TYPE (exp_plus_one) = ASTREF (incrdecr_exp_type);
       ass_stmt = pkl_ast_make_ass_stmt (PKL_PASS_AST,
                                         incrdecr_exp, exp_plus_one);
+      PKL_AST_LOC (ass_stmt) = PKL_AST_LOC (incrdecr);
 
       PKL_AST_INCRDECR_ASS_STMT (incrdecr) = ASTREF (ass_stmt);
       PKL_PASS_RESTART = 1;
@@ -1545,6 +1946,25 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_op_attr)
 }
 PKL_PHASE_END_HANDLER
 
+/* Transform assignment statements to l-value bconc operators into an
+   equivalent sequence of assignments.  */
+
+PKL_PHASE_BEGIN_HANDLER (pkl_trans2_ps_ass_stmt)
+{
+  pkl_ast_node lvalue = PKL_AST_ASS_STMT_LVALUE (PKL_PASS_NODE);
+
+  if (PKL_AST_CODE (lvalue) == PKL_AST_EXP
+      && PKL_AST_EXP_CODE (lvalue) == PKL_AST_OP_BCONC)
+    {
+      pkl_ast_node repl = pkl_ast_handle_bconc_ass_stmt (PKL_PASS_AST,
+                                                         PKL_PASS_NODE);
+      pkl_ast_node_free (PKL_PASS_NODE);
+      PKL_PASS_NODE = ASTREF (repl);
+      PKL_PASS_RESTART = 1;
+    }
+}
+PKL_PHASE_END_HANDLER
+
 struct pkl_phase pkl_phase_trans2 =
   {
    PKL_PHASE_PS_HANDLER (PKL_AST_SRC, pkl_trans_ps_src),
@@ -1558,6 +1978,7 @@ struct pkl_phase pkl_phase_trans2 =
    PKL_PHASE_PS_HANDLER (PKL_AST_STRUCT_REF, pkl_trans2_ps_struct_ref),
    PKL_PHASE_PS_HANDLER (PKL_AST_CAST, pkl_trans2_ps_cast),
    PKL_PHASE_PS_HANDLER (PKL_AST_INCRDECR, pkl_trans2_ps_incrdecr),
+   PKL_PHASE_PS_HANDLER (PKL_AST_ASS_STMT, pkl_trans2_ps_ass_stmt),
    PKL_PHASE_PS_TYPE_HANDLER (PKL_TYPE_OFFSET, pkl_trans2_ps_type_offset),
    PKL_PHASE_PS_HANDLER (PKL_AST_STRUCT_TYPE_FIELD, pkl_trans2_ps_struct_type_field),
    PKL_PHASE_PS_OP_HANDLER (PKL_AST_OP_ATTR, pkl_trans2_ps_op_attr),
@@ -1599,12 +2020,12 @@ PKL_PHASE_BEGIN_HANDLER (pkl_trans3_ps_op_sizeof)
 
     offset_type = pkl_ast_make_offset_type (PKL_PASS_AST,
                                             PKL_AST_TYPE (magnitude),
-                                            unit);
+                                            unit, NULL /* ref_type */);
     PKL_AST_TYPE (offset) = ASTREF (offset_type);
   }
 
   pkl_ast_node_free (PKL_PASS_NODE);
-  PKL_PASS_NODE = offset;
+  PKL_PASS_NODE = ASTREF (offset);
   PKL_PASS_RESTART = 1;
 }
 PKL_PHASE_END_HANDLER
