@@ -1,6 +1,6 @@
 /* pvm-val.c - Values for the PVM.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -38,7 +38,6 @@
 
 static pvm_val string_type;
 static pvm_val void_type;
-static pvm_val any_type;
 
 /* We are currently only supporting a relatively small number of
    integral types, i.e. signed and unsigned types of sizes 1 to 64
@@ -181,6 +180,7 @@ pvm_array_insert (pvm_val arr, pvm_val idx, pvm_val val)
   size_t nelem = PVM_VAL_ULONG (PVM_VAL_ARR_NELEM (arr));
   size_t nallocated = PVM_VAL_ARR_NALLOCATED (arr);
   size_t nelem_to_add = index - nelem + 1;
+  size_t nelem_to_allocate = index >= nallocated ? index - nallocated + 1 : 0;
   size_t val_size = pvm_sizeof (val);
   size_t array_boffset = PVM_VAL_ULONG (PVM_VAL_ARR_OFFSET (arr));
   size_t elem_boffset
@@ -195,9 +195,9 @@ pvm_array_insert (pvm_val arr, pvm_val idx, pvm_val val)
   if (index < nelem)
     return 0;
 
-  /* We have a hard-limit in the number of elements to append, in
+  /* We have a hard-limit in the number of elements to allocate, in
      order to avoid malicious code or harmful bugs.  */
-  if (nelem_to_add > 1024)
+  if (nelem_to_allocate > 1024)
     return 0;
 
   /* Make sure there is enough room in the array for the new elements.
@@ -253,7 +253,7 @@ pvm_array_set (pvm_val arr, pvm_val idx, pvm_val val)
   /* Update the element with the given value.  */
   PVM_VAL_ARR_ELEM_VALUE (arr, index) = val;
 
-  /* Recalculate the bit-offset of all the elemens following the
+  /* Recalculate the bit-offset of all the elements following the
      element just updated.  */
   for (i = index + 1; i < nelem; ++i)
     {
@@ -372,6 +372,28 @@ pvm_ref_struct_cstr (pvm_val sct, const char *name)
   return PVM_NULL;
 }
 
+void
+pvm_ref_set_struct_cstr (pvm_val sct, const char *fname,
+                         pvm_val value)
+{
+  size_t nfields, i;
+  struct pvm_struct_field *fields;
+
+  assert (PVM_IS_SCT (sct));
+
+  nfields = PVM_VAL_ULONG (PVM_VAL_SCT_NFIELDS (sct));
+  fields = PVM_VAL_SCT (sct)->fields;
+
+  for (i = 0; i < nfields; ++i)
+    {
+      if (!PVM_VAL_SCT_FIELD_ABSENT_P (sct, i)
+          && fields[i].name != PVM_NULL
+          && STREQ (PVM_VAL_STR (fields[i].name),
+                    fname))
+        fields[i].value = value;
+    }
+}
+
 pvm_val
 pvm_ref_struct (pvm_val sct, pvm_val name)
 {
@@ -496,28 +518,25 @@ pvm_make_void_type (void)
 }
 
 pvm_val
-pvm_make_any_type (void)
-{
-  return any_type;
-}
-
-pvm_val
-pvm_make_offset_type (pvm_val base_type, pvm_val unit)
+pvm_make_offset_type (pvm_val base_type, pvm_val unit, pvm_val ref_type)
 {
   pvm_val otype = pvm_make_type (PVM_TYPE_OFFSET);
 
   PVM_VAL_TYP_O_BASE_TYPE (otype) = base_type;
   PVM_VAL_TYP_O_UNIT (otype) = unit;
+  PVM_VAL_TYP_O_REF_TYPE (otype) = ref_type;
   return otype;
 }
 
 pvm_val
-pvm_make_array_type (pvm_val type, pvm_val bound)
+pvm_make_array_type (pvm_val type, pvm_val bounder)
 {
   pvm_val atype = pvm_make_type (PVM_TYPE_ARRAY);
 
+  assert (PVM_IS_CLS (bounder));
+
   PVM_VAL_TYP_A_ETYPE (atype) = type;
-  PVM_VAL_TYP_A_BOUND (atype) = bound;
+  PVM_VAL_TYP_A_BOUND (atype) = bounder;
   return atype;
 }
 
@@ -529,6 +548,7 @@ pvm_make_struct_type (pvm_val nfields, pvm_val name,
 
   PVM_VAL_TYP_S_NAME (stype) = name;
   PVM_VAL_TYP_S_NFIELDS (stype) = nfields;
+  PVM_VAL_TYP_S_CONSTRUCTOR (stype) = PVM_NULL;
   PVM_VAL_TYP_S_FNAMES (stype) = fnames;
   PVM_VAL_TYP_S_FTYPES (stype) = ftypes;
 
@@ -549,11 +569,12 @@ pvm_make_closure_type (pvm_val rtype,
 }
 
 pvm_val
-pvm_make_cls (pvm_program program)
+pvm_make_cls (pvm_program program, pvm_val name)
 {
   pvm_val_box box = pvm_make_box (PVM_VAL_TAG_CLS);
   pvm_cls cls = pvm_alloc_cls ();
 
+  cls->name = name;
   cls->program = program;
   cls->entry_point = pvm_program_beginning (program);
   cls->env = NULL; /* This should be set by a PEC instruction before
@@ -564,14 +585,13 @@ pvm_make_cls (pvm_program program)
 }
 
 pvm_val
-pvm_make_offset (pvm_val magnitude, pvm_val unit)
+pvm_make_offset (pvm_val magnitude, pvm_val type)
 {
   pvm_val_box box = pvm_make_box (PVM_VAL_TAG_OFF);
   pvm_off off = pvm_alloc (sizeof (struct pvm_off));
 
-  off->base_type = pvm_typeof (magnitude);
+  off->type = type;
   off->magnitude = magnitude;
-  off->unit = unit;
 
   PVM_VAL_BOX_OFF (box) = off;
   return PVM_BOX (box);
@@ -598,12 +618,14 @@ pvm_val_equal_p (pvm_val val1, pvm_val val2)
     return STREQ (PVM_VAL_STR (val1), PVM_VAL_STR (val2));
   else if (PVM_IS_OFF (val1) && PVM_IS_OFF (val2))
     {
+      pvm_val val1_type = PVM_VAL_OFF_TYPE (val1);
+      pvm_val val2_type = PVM_VAL_OFF_TYPE (val2);
       int pvm_off_mag_equal, pvm_off_unit_equal;
 
       pvm_off_mag_equal = pvm_val_equal_p (PVM_VAL_OFF_MAGNITUDE (val1),
                                            PVM_VAL_OFF_MAGNITUDE (val2));
-      pvm_off_unit_equal = pvm_val_equal_p (PVM_VAL_OFF_UNIT (val1),
-                                            PVM_VAL_OFF_UNIT (val2));
+      pvm_off_unit_equal = pvm_val_equal_p (PVM_VAL_TYP_O_UNIT (val1_type),
+                                            PVM_VAL_TYP_O_UNIT (val2_type));
 
       return pvm_off_mag_equal && pvm_off_unit_equal;
     }
@@ -969,8 +991,11 @@ pvm_sizeof (pvm_val val)
   else if (PVM_IS_CLS (val))
     /* By convention, closure values have size zero.  */
     return 0;
+  else if (val == PVM_NULL)
+    /* By convention, PVM_NULL values have size zero.  */
+    return 0;
 
-  assert (0);
+  PK_UNREACHABLE ();
   return 0;
 }
 
@@ -1120,7 +1145,7 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       uint_fmt = "";
       break;
     default:
-      assert (0);
+      PK_UNREACHABLE ();
       break;
     }
 
@@ -1143,8 +1168,8 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       if (base == 2)
         {
           pk_puts ("0b");
-          pk_print_binary (pk_puts, ulongval, size, /*signed_p*/ 1,
-                           /*use_suffix_p*/ 1);
+          PK_PRINT_BINARY (ulongval, size);
+          pk_puts (PK_INTEGRAL_SUFFIX (size, /*signed_p*/ 1));
         }
       else
         {
@@ -1173,8 +1198,8 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       if (base == 2)
         {
           pk_puts ("0b");
-          pk_print_binary (pk_puts, (uint64_t) uintval, size, /*signed*/ 1,
-                           /*use_suffix_p*/ 1);
+          PK_PRINT_BINARY ((uint64_t) uintval, size);
+          pk_puts (PK_INTEGRAL_SUFFIX (size, /*signed_p*/ 1));
         }
       else
         {
@@ -1203,8 +1228,8 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       if (base == 2)
         {
           pk_puts ("0b");
-          pk_print_binary (pk_puts, ulongval, size, /*signed_p*/ 0,
-                           /*use_suffix_p*/ 1);
+          PK_PRINT_BINARY (ulongval, size);
+          pk_puts (PK_INTEGRAL_SUFFIX (size, /*signed_p*/ 0));
         }
       else
         {
@@ -1226,8 +1251,8 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       if (base == 2)
         {
           pk_puts ("0b");
-          pk_print_binary (pk_puts, uintval, size, /*signed_p*/ 0,
-                           /*use_suffix_p*/ 1);
+          PK_PRINT_BINARY (uintval, size);
+          pk_puts (PK_INTEGRAL_SUFFIX (size, /*signed_p*/ 0));
         }
       else
         {
@@ -1423,15 +1448,15 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
                   pk_puts ("=");
                 }
               PVM_PRINT_VAL_1 (value, ndepth + 1);
-            }
 
-          if (maps && offset != PVM_NULL)
-            {
-              pk_puts (" @ ");
-              pk_term_class ("offset");
-              PVM_PRINT_VAL_1 (offset, ndepth);
-              pk_puts ("#b");
-              pk_term_end_class ("offset");
+              if (maps && offset != PVM_NULL)
+                {
+                  pk_puts (" @ ");
+                  pk_term_class ("offset");
+                  PVM_PRINT_VAL_1 (offset, ndepth);
+                  pk_puts ("#b");
+                  pk_term_end_class ("offset");
+                }
             }
         }
 
@@ -1469,7 +1494,7 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
               case 16: pk_puts ("int16"); break;
               case 32: pk_puts ("int32"); break;
               case 64: pk_puts ("int64"); break;
-              default: assert (0); break;
+              default: PK_UNREACHABLE (); break;
               }
           }
           break;
@@ -1479,11 +1504,6 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
         case PVM_TYPE_VOID:
           pk_puts ("void");
           break;
-        case PVM_TYPE_ANY:
-          pk_term_class ("any");
-          pk_puts ("any");
-          pk_term_end_class ("any");
-          break;
         case PVM_TYPE_ARRAY:
           PVM_PRINT_VAL_1 (PVM_VAL_TYP_A_ETYPE (val), ndepth);
           pk_puts ("[");
@@ -1547,17 +1567,19 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
           break;
           }
         default:
-          assert (0);
+          PK_UNREACHABLE ();
         }
 
       pk_term_end_class ("type");
     }
   else if (PVM_IS_OFF (val))
     {
+      pvm_val val_type = PVM_VAL_OFF_TYPE (val);
+
       pk_term_class ("offset");
       PVM_PRINT_VAL_1 (PVM_VAL_OFF_MAGNITUDE (val), ndepth);
       pk_puts ("#");
-      print_unit_name (PVM_VAL_ULONG (PVM_VAL_OFF_UNIT (val)));
+      print_unit_name (PVM_VAL_ULONG (PVM_VAL_TYP_O_UNIT (val_type)));
       pk_term_end_class ("offset");
     }
   else if (PVM_IS_CLS (val))
@@ -1567,7 +1589,7 @@ pvm_print_val_1 (pvm vm, int depth, int mode, int base, int indent,
       pk_term_end_class ("special");
     }
   else
-    assert (0);
+    PK_UNREACHABLE ();
 }
 
 #undef PVM_PRINT_VAL_1
@@ -1625,16 +1647,17 @@ pvm_typeof (pvm_val val)
   else if (PVM_IS_STR (val))
     type = pvm_make_string_type ();
   else if (PVM_IS_OFF (val))
-    type = pvm_make_offset_type (PVM_VAL_OFF_BASE_TYPE (val),
-                                 PVM_VAL_OFF_UNIT (val));
+    type = PVM_VAL_OFF_TYPE (val);
   else if (PVM_IS_ARR (val))
     type = PVM_VAL_ARR_TYPE (val);
   else if (PVM_IS_SCT (val))
     type = PVM_VAL_SCT_TYPE (val);
   else if (PVM_IS_TYP (val))
     type = val;
+  else if (PVM_IS_CLS (val))
+    type = PVM_NULL;
   else
-    assert (0);
+    PK_UNREACHABLE ();
 
   return type;
 }
@@ -1660,12 +1683,22 @@ pvm_type_equal_p (pvm_val type1, pvm_val type2)
         return (t1_size == t2_size && t1_signed == t2_signed);
       }
     case PVM_TYPE_STRING:
-    case PVM_TYPE_ANY:
     case PVM_TYPE_VOID:
       return 1;
     case PVM_TYPE_ARRAY:
-      return pvm_type_equal_p (PVM_VAL_TYP_A_ETYPE (type1),
-                               PVM_VAL_TYP_A_ETYPE (type2));
+      {
+        pvm_val etype1 = PVM_VAL_TYP_A_ETYPE (type1);
+        pvm_val etype2 = PVM_VAL_TYP_A_ETYPE (type2);
+
+        /* Note that arrays whose elements can be of any
+           type have etype PVM_NULL.  */
+        if (etype1 == PVM_NULL && etype2 == PVM_NULL)
+          return 1;
+        else if (etype1 == PVM_NULL || etype2 == PVM_NULL)
+          return 0;
+        else
+          return pvm_type_equal_p (etype1, etype2);
+      }
     case PVM_TYPE_STRUCT:
       return (STREQ (PVM_VAL_STR (PVM_VAL_TYP_S_NAME (type1)),
                      PVM_VAL_STR (PVM_VAL_TYP_S_NAME (type2))));
@@ -1697,7 +1730,7 @@ pvm_type_equal_p (pvm_val type1, pvm_val type2)
         return 1;
       }
     default:
-      assert (0);
+      PK_UNREACHABLE ();
     }
 }
 
@@ -1801,12 +1834,10 @@ pvm_val_initialize (void)
 
   pvm_alloc_add_gc_roots (&string_type, 1);
   pvm_alloc_add_gc_roots (&void_type, 1);
-  pvm_alloc_add_gc_roots (&any_type, 1);
   pvm_alloc_add_gc_roots (&common_int_types, 65 * 2);
 
   string_type = pvm_make_type (PVM_TYPE_STRING);
   void_type = pvm_make_type (PVM_TYPE_VOID);
-  any_type = pvm_make_type (PVM_TYPE_ANY);
 
   for (i = 0; i < 65; ++i)
     for (j = 0; j < 2; ++j)
@@ -1818,6 +1849,5 @@ pvm_val_finalize (void)
 {
   pvm_alloc_remove_gc_roots (&string_type, 1);
   pvm_alloc_remove_gc_roots (&void_type, 1);
-  pvm_alloc_remove_gc_roots (&any_type, 1);
   pvm_alloc_remove_gc_roots (&common_int_types, 65 * 2);
 }
