@@ -1,6 +1,6 @@
 /* pk-cmd-ios.c - Commands for operating on IO spaces.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,28 +36,80 @@
 #if HAVE_HSERVER
 #  include "pk-hserver.h"
 #endif
+#include "pk-repl.h" /* For poke_completion_function */
+
+/* Get a Poke expression and compile it to get an IO space identifier.
+   Lexical cuckolding is active to allow referring to IO spaces by
+   name with $<...> constructs.  If the expression is invalid or
+   doesn't evaluate to a valid IO space id then return NULL.  */
+
+static pk_ios
+expr_to_ios (const char *expr)
+{
+  pk_val val;
+  pk_val exit_exception = PK_NULL;
+  int ret;
+  pk_ios io;
+
+  pk_set_lexical_cuckolding_p (poke_compiler, 1);
+  ret = pk_compile_expression (poke_compiler, expr, NULL, &val,
+                               &exit_exception);
+  pk_set_lexical_cuckolding_p (poke_compiler, 0);
+
+  if (ret != PK_OK || exit_exception != PK_NULL)
+    {
+      /* The compiler has already printed diagnostics in the
+         terminal.  */
+      if (exit_exception != PK_NULL)
+        poke_handle_exception (exit_exception);
+      return NULL;
+    }
+
+  /* See that VAL is an IO space, i.e. an int<32>.  */
+  if (pk_type_code (pk_typeof (val)) != PK_TYPE_INT
+      || pk_int_size (val) != 32)
+    {
+      pk_term_class ("error");
+      pk_puts (_("error:"));
+      pk_term_end_class ("error");
+      pk_puts (" please specify a valid IO space\n");
+      return NULL;
+    }
+
+  io = pk_ios_search_by_id (poke_compiler, pk_int_value (val));
+  if (io == NULL)
+    {
+      pk_term_class ("error");
+      pk_puts (_("error:"));
+      pk_term_end_class ("error");
+      pk_printf (" no such IO space with id %" PRIx64 "\n", pk_int_value (val));
+      return NULL;
+    }
+
+  return io;
+}
 
 static int
 pk_cmd_ios (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
 {
-  /* ios #ID */
-
-  int io_id;
+  /* ios EXPR */
   pk_ios io;
 
   assert (argc == 2);
-  assert (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_TAG);
 
-  io_id = PK_CMD_ARG_TAG (argv[1]);
-  io = pk_ios_search_by_id (poke_compiler, io_id);
-  if (io == NULL)
+  if (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_NULL)
+    /* Do nothing.  */
+    return 0;
+  else
     {
-      pk_printf (_("No IOS with tag #%d\n"), io_id);
-      return 0;
+      assert (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_STR);
+
+      io = expr_to_ios (PK_CMD_ARG_STR (argv[1]));
+      if (io == NULL)
+        return 0;
     }
 
   pk_ios_set_cur (poke_compiler, io);
-
   if (poke_interactive_p && !poke_quiet_p)
     pk_printf (_("The current IOS is now `%s'.\n"),
                pk_ios_handler (pk_ios_cur (poke_compiler)));
@@ -68,28 +120,31 @@ static int
 pk_cmd_sub (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
 {
   char *handler, *name;
-  int ios;
+  pk_ios ios;
   uint64_t base, size;
 
   assert (argc == 5);
 
   /* Collect and validate arguments.  */
 
-  assert (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_TAG);
-  ios = PK_CMD_ARG_TAG (argv[1]);
+  assert (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_STR);
+  ios = expr_to_ios (PK_CMD_ARG_STR (argv[1]));
+  if (ios == NULL)
+    return 0;
 
-  assert (PK_CMD_ARG_TYPE (argv[2]) == PK_CMD_ARG_INT);
-  base = PK_CMD_ARG_INT (argv[2]);
+  assert (PK_CMD_ARG_TYPE (argv[2]) == PK_CMD_ARG_UINT);
+  base = PK_CMD_ARG_UINT (argv[2]);
 
-  assert (PK_CMD_ARG_TYPE (argv[3]) == PK_CMD_ARG_INT);
-  size = PK_CMD_ARG_INT (argv[3]);
+  assert (PK_CMD_ARG_TYPE (argv[3]) == PK_CMD_ARG_UINT);
+  size = PK_CMD_ARG_UINT (argv[3]);
 
-  assert (PK_CMD_ARG_TYPE (argv[4]) == PK_CMD_ARG_STR);
-  name = PK_CMD_ARG_STR (argv[4]);
+  name = (PK_CMD_ARG_TYPE (argv[4]) == PK_CMD_ARG_STR
+          ? PK_CMD_ARG_STR (argv[4])
+          : NULL);
 
   /* Build the handler.  */
   if (asprintf (&handler, "sub://%d/0x%" PRIx64 "/0x%" PRIx64 "/%s",
-                ios, base, size, name) == -1)
+                pk_ios_get_id (ios), base, size, name ? name : "") == -1)
     return 0;
 
   /* Open the IOS.  */
@@ -135,7 +190,7 @@ pk_cmd_proc (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
     }
   free (handler);
 
-  /* Allright, now open sub spaces for the process' maps, if
+  /* All right, now open sub spaces for the process' maps, if
      requested.  */
   if (uflags & PK_PROC_F_MAPS || uflags & PK_PROC_F_MAPS_ALL)
     pk_open_proc_maps (ios_id, pid, uflags & PK_PROC_F_MAPS_ALL);
@@ -171,9 +226,9 @@ pk_cmd_file (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
     /* /c has no effect if the file exists.  */
     create_p = 0;
 
-  if (pk_ios_search (poke_compiler, filename) != NULL)
+  if (pk_ios_search (poke_compiler, filename, PK_IOS_SEARCH_F_EXACT) != NULL)
     {
-      printf (_("File %s already opened.  Use `.ios #N' to switch.\n"),
+      printf (_("File %s already opened.  Use `.ios IOS' to switch.\n"),
               filename);
       return 0;
     }
@@ -223,32 +278,27 @@ pk_cmd_close (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
 {
   /* close [#ID]  */
   pk_ios io;
-  int io_id;
+  const char *expr;
 
   assert (argc == 2);
+  assert (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_STR);
+  expr = PK_CMD_ARG_STR (argv[1]);
 
-  if (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_NULL)
-    {
-      io = pk_ios_cur (poke_compiler);
-      io_id = pk_ios_get_id (io);
-    }
+  if (*expr == '\0')
+    io = pk_ios_cur (poke_compiler);
   else
     {
-      io_id = PK_CMD_ARG_TAG (argv[1]);
-
-      io = pk_ios_search_by_id (poke_compiler, io_id);
+      io = expr_to_ios (expr);
       if (io == NULL)
-        {
-          pk_printf (_("No such IO space: #%d\n"), io_id);
-          return 0;
-        }
+        return 0;
     }
 
   pk_ios_close (poke_compiler, io);
 
   /* All right, now we want to close all the open IOS which are subs
      of the space we just closed.  */
-  pk_ios_map (poke_compiler, close_if_sub_of, (void *) (intptr_t) io_id);
+  pk_ios_map (poke_compiler, close_if_sub_of,
+              (void *) (intptr_t) pk_ios_get_id (io));
 
   return 1;
 }
@@ -266,7 +316,7 @@ print_info_ios (pk_ios io, void *data)
   {
     char *tag;
 
-    asprintf (&tag, "%s#%d",
+    asprintf (&tag, "%s%d",
               io == pk_ios_cur (poke_compiler) ? "* " : "  ",
               pk_ios_get_id (io));
     pk_table_column (table, tag);
@@ -290,7 +340,7 @@ print_info_ios (pk_ios io, void *data)
       asprintf (&string, "0x%08" PRIx64 "#B", bias / 8);
     else
       asprintf (&string, "0x%08" PRIx64 "#b", bias);
-    pk_table_column_cl (table, string, "offset");
+    pk_table_column (table, string);
     free (string);
   }
 
@@ -298,7 +348,7 @@ print_info_ios (pk_ios io, void *data)
   {
     char *size;
     asprintf (&size, "0x%08" PRIx64 "#B", pk_ios_size (io));
-    pk_table_column_cl (table, size, "offset");
+    pk_table_column (table, size);
     free (size);
   }
 
@@ -309,7 +359,7 @@ print_info_ios (pk_ios io, void *data)
     char *cmd;
     char *hyperlink;
 
-    asprintf (&cmd, ".ios #%d", pk_ios_get_id (io));
+    asprintf (&cmd, ".ios %d", pk_ios_get_id (io));
     hyperlink = pk_hserver_make_hyperlink ('e', cmd, PK_NULL);
     free (cmd);
 
@@ -327,7 +377,7 @@ print_info_ios (pk_ios io, void *data)
       char *cmd;
       char *hyperlink;
 
-      asprintf (&cmd, ".close #%d", pk_ios_get_id (io));
+      asprintf (&cmd, ".close %d", pk_ios_get_id (io));
       hyperlink = pk_hserver_make_hyperlink ('e', cmd, PK_NULL);
       free (cmd);
 
@@ -454,7 +504,7 @@ pk_cmd_mem (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
   const char *arg_str = PK_CMD_ARG_STR (argv[1]);
   char *mem_name = NULL;
 
-  /* If the user didnt' specify a name for the memory IOS, search for
+  /* If the user didn't specify a name for the memory IOS, search for
      the first available *N*.  */
   if (*arg_str == '\0')
     {
@@ -466,7 +516,8 @@ pk_cmd_mem (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
           if (asprintf (&mem_name, "*%d*", i) == -1)
             pk_fatal (_("out of memory"));
 
-          if (pk_ios_search (poke_compiler, mem_name) == NULL)
+          if (pk_ios_search (poke_compiler, mem_name, PK_IOS_SEARCH_F_EXACT)
+              == NULL)
             break;
         }
     }
@@ -476,9 +527,9 @@ pk_cmd_mem (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
         pk_fatal (_("out of memory"));
     }
 
-  if (pk_ios_search (poke_compiler, mem_name) != NULL)
+  if (pk_ios_search (poke_compiler, mem_name, PK_IOS_SEARCH_F_EXACT) != NULL)
     {
-      printf (_("Buffer %s already opened.  Use `.ios #N' to switch.\n"),
+      printf (_("Buffer %s already opened.  Use `.ios IOS' to switch.\n"),
               mem_name);
       free (mem_name);
       return 0;
@@ -513,9 +564,9 @@ pk_cmd_nbd (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
   const char *arg_str = PK_CMD_ARG_STR (argv[1]);
   char *nbd_name = xstrdup (arg_str);
 
-  if (pk_ios_search (poke_compiler, nbd_name) != NULL)
+  if (pk_ios_search (poke_compiler, nbd_name, PK_IOS_SEARCH_F_EXACT) != NULL)
     {
-      printf (_("Buffer %s already opened.  Use `.ios #N' to switch.\n"),
+      printf (_("Buffer %s already opened.  Use `.ios IOS' to switch.\n"),
               nbd_name);
       free (nbd_name);
       return 0;
@@ -536,44 +587,94 @@ pk_cmd_nbd (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
 }
 #endif /* HAVE_LIBNBD */
 
-static char *
-ios_completion_function (const char *x, int state)
+#ifdef HAVE_MMAP
+static int
+pk_cmd_mmap (int argc, struct pk_cmd_arg argv[], uint64_t uflags)
 {
-  return pk_ios_completion_function (poke_compiler, x, state);
+  /* mmap FILENAME BASE SIZE */
+  uint64_t base, size;
+  char *handler;
+  int ios_id;
+
+  assert (argc == 4);
+  assert (PK_CMD_ARG_TYPE (argv[1]) == PK_CMD_ARG_STR);
+
+  /* Create a new IO space.  */
+  const char *filename = PK_CMD_ARG_STR (argv[1]);
+
+  assert (PK_CMD_ARG_TYPE (argv[2]) == PK_CMD_ARG_UINT);
+  base = PK_CMD_ARG_UINT (argv[2]);
+
+  assert (PK_CMD_ARG_TYPE (argv[3]) == PK_CMD_ARG_UINT);
+  size = PK_CMD_ARG_UINT (argv[3]);
+
+  if (access (filename, F_OK) == 0)
+
+  if (pk_ios_search (poke_compiler, filename, PK_IOS_SEARCH_F_EXACT) != NULL)
+    {
+      printf (_("File %s already opened.  Use `.ios IOS' to switch.\n"),
+              filename);
+      return 0;
+    }
+
+  /* Build the handler for the mmap IOS.  */
+  if (asprintf (&handler, "mmap://0x%" PRIx64 "/0x%" PRIx64 "/%s",
+                           base, size, filename) == -1)
+    return 0;
+
+  /* Open the IOS.  */
+  ios_id = pk_ios_open (poke_compiler, handler, 0, 1);
+  if (ios_id == PK_IOS_NOID)
+    {
+      pk_printf (_("Error creating mmap IOS %s\n"), handler);
+      free (handler);
+      return 0;
+    }
+  free (handler);
+
+  return 1;
 }
+#endif /* HAVE_MMAP */
 
 const struct pk_cmd ios_cmd =
-  {"ios", "t", "", 0, NULL, NULL, pk_cmd_ios, "ios #ID", ios_completion_function};
+  {"ios", "?s", "", 0, NULL, NULL, pk_cmd_ios, ".ios IOS", poke_completion_function};
 
 const struct pk_cmd file_cmd =
-  {"file", "f", PK_FILE_UFLAGS, 0, NULL, NULL, pk_cmd_file, "file FILE-NAME",
+  {"file", "f", PK_FILE_UFLAGS, 0, NULL, NULL, pk_cmd_file, ".file FILE-NAME",
    rl_filename_completion_function};
 
 const struct pk_cmd proc_cmd =
-  {"proc", "i", PK_PROC_UFLAGS, 0, NULL, NULL, pk_cmd_proc, "proc PID", NULL};
+  {"proc", "i", PK_PROC_UFLAGS, 0, NULL, NULL, pk_cmd_proc, ".proc PID", NULL};
 
 const struct pk_cmd sub_cmd =
-  {"sub", "t,i,i,?s", "", 0, NULL, NULL, pk_cmd_sub, "sub IOS, BASE, SIZE, [NAME]", NULL};
+  {"sub", "s,u,u,?s", "", 0, NULL, NULL, pk_cmd_sub, ".sub IOS, BASE, SIZE, [NAME]",
+   poke_completion_function};
 
 const struct pk_cmd mem_cmd =
-  {"mem", "s", "", 0, NULL, NULL, pk_cmd_mem, "mem NAME", NULL};
+  {"mem", "s", "", 0, NULL, NULL, pk_cmd_mem, ".mem NAME", NULL};
 
 #ifdef HAVE_LIBNBD
 const struct pk_cmd nbd_cmd =
-  {"nbd", "s", "", 0, NULL, NULL, pk_cmd_nbd, "nbd URI", NULL};
+  {"nbd", "s", "", 0, NULL, NULL, pk_cmd_nbd, ".nbd URI", NULL};
+#endif
+
+#ifdef HAVE_MMAP
+const struct pk_cmd mmap_cmd =
+  {"mmap", "s,u,u", "", 0, NULL, NULL, pk_cmd_mmap, ".mmap FILE-NAME, BASE, SIZE",
+   rl_filename_completion_function};
 #endif
 
 const struct pk_cmd close_cmd =
-  {"close", "?t", "", PK_CMD_F_REQ_IO, NULL, NULL, pk_cmd_close,
-   "close [#ID]", ios_completion_function};
+  {"close", "s", "", PK_CMD_F_REQ_IO, NULL, NULL, pk_cmd_close,
+   ".close [IOS]", poke_completion_function};
 
 const struct pk_cmd info_ios_cmd =
-  {"ios", "", "", 0, NULL, NULL, pk_cmd_info_ios, "info ios", NULL};
+  {"ios", "", "", 0, NULL, NULL, pk_cmd_info_ios, ".info ios", NULL};
 
 const struct pk_cmd load_cmd =
-  {"load", "f", "", 0, NULL, NULL, pk_cmd_load_file, "load FILE-NAME",
+  {"load", "f", "", 0, NULL, NULL, pk_cmd_load_file, ".load FILE-NAME",
    rl_filename_completion_function};
 
 const struct pk_cmd source_cmd =
   {"source", "f", "", 0, NULL, NULL, pk_cmd_source_file,
-   "source FILE-NAME", rl_filename_completion_function};
+   ".source FILE-NAME", rl_filename_completion_function};
