@@ -1,6 +1,6 @@
 /* poke.c - Interactive editor for binary files.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023, 2024 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -26,7 +26,10 @@
 #include <locale.h>
 #include <textstyle.h>
 #include <assert.h>
+#ifndef _WIN32
 #include <arpa/inet.h> /* For htonl */
+#endif
+#include <xalloc.h>
 
 #ifdef HAVE_HSERVER
 #  include "pk-hserver.h"
@@ -36,9 +39,9 @@
 #include "pk-cmd.h"
 #include "pk-repl.h"
 #include "pk-utils.h"
-#include "pk-mi.h"
 #include "pk-map.h"
 #include "pk-ios.h"
+#include "configmake.h"
 
 /* poke can be run either interactively (from a tty) or in batch mode.
    The following predicate records this.  */
@@ -51,29 +54,22 @@ int poke_interactive_p;
 int poke_hserver_p;
 #endif
 
-/* The following global indicates whether the MI shall be used.  */
-int poke_mi_p;
-
-#ifdef POKE_MI
-/* This build of poke supports mi if the user wants it.  */
-static const int mi_supported_p = 1;
-#else
-/* This build of poke does not support mi.  */
-static const int mi_supported_p = 0;
-int pk_mi (void) {assert (0); return 0;}
-#endif
-
 /* The following global indicates whether poke should be as terse as
    possible in its output.  This is useful when running poke from
    other programs.  */
 
 int poke_quiet_p;
 
-/* The following global contains the directory holding the program's
+/* The following global contains the directory holding libpoke's
    architecture independent files, such as scripts.  */
 
 char *poke_datadir;
 
+/* The following global contains the directory holding the program's
+   architecture independent files, such as scripts.  */
+
+char *poke_appdir;
+
 /* The following global contains the directory holding the program's
    info files.  */
 
@@ -104,7 +100,7 @@ char *poke_mapsdir;
 char *poke_docdir;
 
 /* The following global determines whether auto-maps shall be
-   acknowleged when loading files by default.  Defaults to 1, but can
+   acknowledged when loading files by default.  Defaults to 1, but can
    be changed to 0 using the --no-auto-map command-line option.  */
 
 int poke_default_auto_map_p = 1;
@@ -114,6 +110,11 @@ int poke_default_auto_map_p = 1;
 
 int poke_no_hserver_arg = 0;
 
+/* The following global determines whether the user specified to not
+   load "standard" types in the Poke incremental compiler.  */
+
+int poke_no_stdtypes_arg = 0;
+
 /* This is used by commands to indicate to the REPL that it must
    exit.  */
 
@@ -144,9 +145,9 @@ enum
   STYLE_ARG,
   STYLE_DARK_ARG,
   STYLE_BRIGHT_ARG,
-  MI_ARG,
   NO_AUTO_MAP_ARG,
-  NO_HSERVER_ARG
+  NO_HSERVER_ARG,
+  NO_STDTYPES_ARG
 };
 
 static const struct option long_options[] =
@@ -162,67 +163,73 @@ static const struct option long_options[] =
   {"style", required_argument, NULL, STYLE_ARG},
   {"style-dark", no_argument, NULL, STYLE_DARK_ARG},
   {"style-bright", no_argument, NULL, STYLE_BRIGHT_ARG},
-  {"mi", no_argument, NULL, MI_ARG},
   {"no-auto-map", no_argument, NULL, NO_AUTO_MAP_ARG},
   {"no-hserver", no_argument, NULL, NO_HSERVER_ARG},
+  {"no-stdtypes", no_argument, NULL, NO_STDTYPES_ARG},
   {NULL, 0, NULL, 0},
 };
 
+/* String passed to getopt_long.  */
+static const char *const poke_getopt_string = "ql:c:s:L:";
+
 static void
 print_help (void)
 {
   /* TRANSLATORS: --help output, GNU poke synopsis.
      no-wrap */
-  pk_puts (_("Usage: poke [OPTION]... [FILE]\n"));
+  puts (_("Usage: poke [OPTION]... [FILE]"));
 
   /* TRANSLATORS: --help output, GNU poke summary.
      no-wrap */
-  pk_puts (_("Interactive editor for binary files.\n"));
+  puts (_("Interactive editor for binary files."));
 
-  pk_puts ("\n");
+  puts ("");
   /* TRANSLATORS: --help output, GNU poke arguments.
      no-wrap */
-  pk_puts (_("  -l, --load=FILE                     load the given pickle at startup\n"));
-  pk_puts (_("  -L FILE                             load the given pickle and exit\n"));
+  puts (_("  -l, --load=FILE                     load the given pickle at startup"));
+  puts (_("  -L FILE                             load the given pickle and exit"));
 
-  pk_puts ("\n");
+  puts ("");
 
   /* TRANSLATORS: --help output, GNU poke arguments.
      no-wrap */
-  pk_puts (_("Commanding poke from the command line:\n"));
-  pk_puts (_("  -c, --command=CMD                   execute the given command\n"));
-  pk_puts (_("  -s, --source=FILE                   execute commands from FILE\n"));
+  puts (_("Commanding poke from the command line:"));
+  puts (_("  -c, --command=CMD                   execute the given command"));
+  puts (_("  -s, --source=FILE                   execute commands from FILE"));
 
-  pk_puts ("\n");
-  pk_puts (_("Styling text output:\n"));
-  pk_puts (_("      --color=(yes|no|auto|html|test) emit styled output\n"));
-  pk_puts (_("      --style=STYLE_FILE              style file to use when styling\n"));
-  pk_puts (_("      --style-dark                    use default style for dark backgrounds\n"));
-  pk_puts (_("      --style-bright                  use default style for bright backgrounds\n"));
+  puts ("");
+  puts (_("Styling text output:"));
+  puts (_("      --color=(yes|no|auto|html|test) emit styled output"));
+  puts (_("      --style=STYLE_FILE              style file to use when styling"));
+  puts (_("      --style-dark                    use default style for dark backgrounds"));
+  puts (_("      --style-bright                  use default style for bright backgrounds"));
 
-  pk_puts ("\n");
-  pk_puts (_("Machine interface:\n"));
-  pk_puts (_("      --mi                            use the MI in stdin/stdout\n"));
-
-  pk_puts ("\n");
+  puts ("");
   /* TRANSLATORS: --help output, less used GNU poke arguments.
      no-wrap */
-  pk_puts (_("  -q, --no-init-file                  do not load an init file\n"));
-  pk_puts (_("      --no-auto-map                   disable auto-map\n"));
+  puts (_("  -q, --no-init-file                  do not load an init file"));
+  puts (_("      --no-auto-map                   disable auto-map"));
 #if HAVE_HSERVER
-  pk_puts (_("      --no-hserver                    do not run the hyperlinks server\n"));
+  puts (_("      --no-hserver                    do not run the hyperlinks server"));
 #endif
-  pk_puts (_("      --quiet                         be as terse as possible\n"));
-  pk_puts (_("      --help                          print a help message and exit\n"));
-  pk_puts (_("      --version                       show version and exit\n"));
+  puts (_("      --no-stdtypes                   do not define standard types"));
+  puts (_("      --quiet                         be as terse as possible"));
+  puts (_("      --help                          print a help message and exit"));
+  puts (_("      --version                       show version and exit"));
 
-  pk_puts ("\n");
+  puts ("");
+  puts (_("The following environment variables, if set, are used by poke:"));
+  puts ("");
+  puts (_("      POKE_LOAD_PATH                  List of file paths separated"));
+  puts (_("                                      by colon characters (:) which"));
+  puts (_("                                      is prepended to the load_path"));
+  puts (_("                                      when poke starts"));
   /* TRANSLATORS: --help output 5+ (reports)
      TRANSLATORS: the placeholder indicates the bug-reporting address
      for this application.  Please add _another line_ with the
      address for translation bugs.
      no-wrap */
-  pk_printf (_("\
+  printf (_("\
 Report bugs in the bug tracker at\n\
   <%s>\n\
   or by email to <%s>.\n"), PACKAGE_BUGZILLA, PACKAGE_BUGREPORT);
@@ -230,8 +237,8 @@ Report bugs in the bug tracker at\n\
   printf (_("Report %s bugs to: %s\n"), PACKAGE_PACKAGER,
           PACKAGE_PACKAGER_BUG_REPORTS);
 #endif
-  pk_printf (_("%s home page: <%s>\n"), PACKAGE_NAME, PACKAGE_URL);
-  pk_printf (_("General help using GNU software: %s\n"), "<http://www.gnu.org/gethelp/>");
+  printf (_("%s home page: <%s>\n"), PACKAGE_NAME, PACKAGE_URL);
+  printf (_("General help using GNU software: %s\n"), "<http://www.gnu.org/gethelp/>");
 }
 
 void
@@ -239,45 +246,38 @@ pk_print_version (int hand_p)
 {
   if (hand_p)
     {
-      pk_term_class ("logo");
-      pk_puts ("     _____\n");
-      pk_puts (" ---'   __\\_______\n");
-      pk_printf ("            ______)  GNU poke %s\n", VERSION);
-      pk_puts ("            __)\n");
-      pk_puts ("           __)\n");
-      pk_puts (" ---._______)\n");
-      pk_term_end_class ("logo");
+      puts ("     _____");
+      puts (" ---'   __\\_______");
+      printf ("            ______)  GNU poke %s\n", VERSION);
+      puts ("            __)");
+      puts ("           __)");
+      puts (" ---._______)");
       /* xgettext: no-wrap */
-      pk_puts ("\n");
+      puts ("");
     }
   else
-    pk_printf ("GNU poke %s\n\n", VERSION);
+    printf ("GNU poke %s\n\n", VERSION);
 
   /* It is important to separate the year from the rest of the message,
      as done here, to avoid having to retranslate the message when a new
      year comes around.  */
-  pk_term_class ("copyright");
   /* TRANSLATORS:
      If your target locale supports it, you can translate (C) to the
      copyright symbol (U+00A9 in Unicode), but there is no obligation
      to do this.  In other cases it's probably best to leave it untranslated.  */
-  pk_printf (_("\
+  printf (_("\
 %s (C) %s The poke authors.\n\
-License GPLv3+: GNU GPL version 3 or later"), "Copyright", "2019-2022");
-  pk_term_hyperlink ("http://gnu.org/licenses/gpl.html", NULL);
-  pk_puts (" <http://gnu.org/licenses/gpl.html>");
-  pk_term_end_hyperlink ();
-  pk_puts (".\n\
+License GPLv3+: GNU GPL version 3 or later"), "Copyright", "2024");
+  puts (".\n\
 This is free software: you are free to change and redistribute it.\n\
-There is NO WARRANTY, to the extent permitted by law.\n");
-  pk_term_end_class ("copyright");
+There is NO WARRANTY, to the extent permitted by law.");
 
-    pk_printf (_("\
+  printf (_("\
 \nPowered by Jitter %s."), JITTER_VERSION);
 
-    pk_puts (_("\
+  puts (_("\
 \n\
-Perpetrated by Jose E. Marchesi.\n"));
+Perpetrated by Jose E. Marchesi."));
 
 }
 
@@ -289,6 +289,7 @@ finalize (void)
     pk_hserver_shutdown ();
 #endif
   pk_cmd_shutdown ();
+  pk_ios_shutdown ();
   pk_map_shutdown ();
   pk_compiler_free (poke_compiler);
   pk_term_shutdown ();
@@ -298,18 +299,18 @@ finalize (void)
 
 static struct pk_term_if poke_term_if =
   {
-    .flush_fn = pk_term_flush,
-    .puts_fn = pk_puts,
-    .printf_fn = pk_printf,
-    .indent_fn = pk_term_indent,
-    .class_fn = pk_term_class,
-    .end_class_fn = pk_term_end_class,
-    .hyperlink_fn = pk_term_hyperlink,
-    .end_hyperlink_fn = pk_term_end_hyperlink,
-    .get_color_fn = pk_term_get_color,
-    .set_color_fn = pk_term_set_color,
-    .get_bgcolor_fn = pk_term_get_bgcolor,
-    .set_bgcolor_fn = pk_term_set_bgcolor,
+    .flush_fn = pk_term_flush_1,
+    .puts_fn = pk_puts_1,
+    .printf_fn = pk_printf_1,
+    .indent_fn = pk_term_indent_1,
+    .class_fn = pk_term_class_1,
+    .end_class_fn = pk_term_end_class_1,
+    .hyperlink_fn = pk_term_hyperlink_1,
+    .end_hyperlink_fn = pk_term_end_hyperlink_1,
+    .get_color_fn = pk_term_get_color_1,
+    .set_color_fn = pk_term_set_color_1,
+    .get_bgcolor_fn = pk_term_get_bgcolor_1,
+    .set_bgcolor_fn = pk_term_set_bgcolor_1,
   };
 
 const char *
@@ -334,7 +335,7 @@ void
 pk_set_var_int (const char *name, int value)
 {
   pk_decl_set_val (poke_compiler, name,
-                   pk_make_int (value, 32));
+                   pk_make_int (poke_compiler, value, 32));
 }
 
 static void
@@ -343,22 +344,30 @@ set_script_args (int argc, char *argv[])
   int i, nargs;
   uint64_t index;
   pk_val argv_array;
+  pk_val array_type_bounder = pk_decl_val (poke_compiler, "pk_bounder_unbounded");
 
   /* Look for -L SCRIPT */
   for (i = 0; i < argc; ++i)
-    if (STREQ (argv[i], "-L"))
-      break;
+    {
+      char *optarg = argv[i];
+
+      if (strlen (optarg) >= 2
+          && optarg[0] == '-'
+          && optarg[1] == 'L')
+        break;
+    }
 
   /* Any argument after SCRIPT is an argument for the script.  */
   i = i + 2;
   nargs = argc - i;
-  argv_array = pk_make_array (pk_make_uint (nargs, 64),
-                              pk_make_array_type (pk_make_string_type (),
-                                                  PK_NULL /* bound */));
+  argv_array = pk_make_array (
+      poke_compiler, pk_make_uint (poke_compiler, nargs, 64),
+      pk_make_array_type (poke_compiler, pk_make_string_type (poke_compiler),
+                          array_type_bounder));
 
   for (index = 0; i < argc; ++i, ++index)
     pk_array_insert_elem (argv_array, index,
-                          pk_make_string (argv[i]));
+                          pk_make_string (poke_compiler, argv[i]));
 
   pk_defvar (poke_compiler, "argv", argv_array);
 }
@@ -371,24 +380,20 @@ parse_args_1 (int argc, char *argv[])
 
   while ((ret = getopt_long (argc,
                              argv,
-                             "ql:c:s:L:",
+                             poke_getopt_string,
                              long_options,
                              NULL)) != -1)
     {
       c = ret;
       switch (c)
         {
-        case MI_ARG:
-          if (!mi_supported_p)
-            {
-              fputs (_("MI is not built into this instance of poke\n"),
-                     stderr);
-              exit (EXIT_FAILURE);
-            }
-          else
-            {
-              poke_mi_p = 1;
-            }
+        case HELP_ARG:
+          print_help ();
+          exit (EXIT_SUCCESS);
+          break;
+        case VERSION_ARG:
+          pk_print_version (0 /* hand_p */);
+          exit (EXIT_SUCCESS);
           break;
         case 'L':
           poke_interactive_p = 0;
@@ -399,11 +404,32 @@ parse_args_1 (int argc, char *argv[])
         case NO_HSERVER_ARG:
           poke_no_hserver_arg = 1;
           break;
+        case NO_STDTYPES_ARG:
+          poke_no_stdtypes_arg = 1;
+          break;
         case 'q':
         case NO_INIT_FILE_ARG:
           poke_load_init_file = 0;
           break;
+        case '?':
+          exit (EXIT_FAILURE);
+          break;
+        case QUIET_ARG:
+        case LOAD_ARG:
+        case LOAD_AND_EXIT_ARG:
+        case CMD_ARG:
+        case SOURCE_ARG:
+        case COLOR_ARG:
+        case STYLE_ARG:
+        case STYLE_BRIGHT_ARG:
+        case STYLE_DARK_ARG:
+        case 'c':
+        case 'l':
+        case 's':
+          /* Handled in parse_args_2.  */
+          break;
         default:
+          assert (!"Did you forget to add a CASE for a flag in parse_args_1?");
           break;
         }
     }
@@ -418,21 +444,13 @@ parse_args_2 (int argc, char *argv[])
   optind = 1;
   while ((ret = getopt_long (argc,
                              argv,
-                             "ql:c:s:L:",
+                             poke_getopt_string,
                              long_options,
                              NULL)) != -1)
     {
       c = ret;
       switch (c)
         {
-        case HELP_ARG:
-          print_help ();
-          goto exit_success;
-          break;
-        case VERSION_ARG:
-          pk_print_version (0 /* hand_p */);
-          goto exit_success;
-          break;
         case QUIET_ARG:
           poke_quiet_p = 1;
           pk_set_quiet_p (poke_compiler, 1);
@@ -490,13 +508,20 @@ parse_args_2 (int argc, char *argv[])
             exit (exit_status);
             break;
           }
-        case MI_ARG:
         case NO_AUTO_MAP_ARG:
         case NO_HSERVER_ARG:
+        case NO_STDTYPES_ARG:
         case 'q':
         case NO_INIT_FILE_ARG:
           /* These are handled in parse_args_1.  */
           break;
+        case HELP_ARG:
+        case VERSION_ARG:
+          /* Handled in parse_args_1.  */
+          break;
+        case LOAD_AND_EXIT_ARG:
+          /* Handled in set_script_args.  */
+          break;
           /* libtextstyle arguments are handled in pk-term.c, not
              here.   */
         case COLOR_ARG:
@@ -504,8 +529,11 @@ parse_args_2 (int argc, char *argv[])
         case STYLE_DARK_ARG:
         case STYLE_BRIGHT_ARG:
           break;
-        default:
+        case '?':
           goto exit_failure;
+        default:
+          assert (!"Did you forget to add a case for a flag in parse_args_2?");
+          break;
         }
     }
 
@@ -517,7 +545,7 @@ parse_args_2 (int argc, char *argv[])
                        filename, 0, 1 /* set_cur_p */) == PK_IOS_NOID)
         {
           if (!poke_quiet_p)
-            pk_printf (_("cannot open file %s\n"), filename);
+            printf (_("cannot open file %s\n"), filename);
           goto exit_failure;
         }
 
@@ -558,12 +586,16 @@ initialize (int argc, char *argv[])
   if (poke_datadir == NULL)
     poke_datadir = PKGDATADIR;
 
+  poke_appdir = getenv ("POKEAPPDIR");
+  if (poke_appdir == NULL)
+    {
+      poke_appdir = pk_str_concat (poke_datadir, "/poke", NULL);
+      pk_assert_alloc (poke_appdir);
+    }
+
   poke_picklesdir = getenv ("POKEPICKLESDIR");
   if (poke_picklesdir == NULL)
-    {
-      poke_picklesdir = pk_str_concat (poke_datadir, "/pickles", NULL);
-      pk_assert_alloc (poke_picklesdir);
-    }
+    poke_picklesdir = "%DATADIR%/pickles";
 
   poke_mapsdir = getenv ("POKEMAPSDIR");
   if (poke_mapsdir == NULL)
@@ -574,7 +606,7 @@ initialize (int argc, char *argv[])
 
   poke_docdir = getenv ("POKEDOCDIR");
   if (poke_docdir == NULL)
-    poke_docdir = poke_datadir;
+    poke_docdir = poke_appdir;
 
   poke_infodir = getenv ("POKEINFODIR");
   if (poke_infodir == NULL)
@@ -584,24 +616,63 @@ initialize (int argc, char *argv[])
   pk_term_init (argc, argv);
 
   /* Initialize the poke incremental compiler.  */
-  poke_compiler = pk_compiler_new (&poke_term_if);
+  poke_compiler = pk_compiler_new_with_flags (&poke_term_if,
+                                              poke_no_stdtypes_arg ? PK_F_NOSTDTYPES : 0);
   if (poke_compiler == NULL)
     pk_fatal ("creating the incremental compiler");
 
   /* Make poke_interactive_p available to poke programs.  */
   if (pk_defvar (poke_compiler, "poke_interactive_p",
-                 pk_make_int (poke_interactive_p, 32)) == PK_ERROR)
+                 pk_make_int (poke_compiler, poke_interactive_p, 32))
+      == PK_ERROR)
     pk_fatal ("defining poke_interactive_p");
 
-  /* The Poke hyperlinks facilities must be loaded before poke.pk and
-     the cmd subsystem.  This is done even if the hserver is
-     disabled.  */
-  if (!pk_load (poke_compiler, "pk-hserver"))
-    pk_fatal ("unable to load the pk-hserver module");
+  /* Add the directories where poke installs .pk files to the
+     compiler's load path.  */
+  {
+    pk_val load_path = pk_decl_val (poke_compiler, "load_path");
+    char *newpaths;
+    const char *env_load_path = getenv ("POKE_LOAD_PATH");
+    char *user_load_path = NULL;
 
-  /* Load poke.pk  */
-  if (!pk_load (poke_compiler, "poke"))
-    pk_fatal ("unable to load the poke module");
+    if (env_load_path)
+      {
+        user_load_path = pk_str_concat (env_load_path, ":", NULL);
+        pk_assert_alloc (user_load_path);
+      }
+
+    assert (load_path != PK_NULL);
+    newpaths = pk_str_concat (user_load_path ? user_load_path : "",
+                              pk_string_str (load_path),
+                              ":",
+                              poke_appdir,
+                              ":",
+                              poke_picklesdir,
+                              NULL);
+    pk_assert_alloc (newpaths);
+
+    pk_decl_set_val (poke_compiler,
+                     "load_path", pk_make_string (poke_compiler, newpaths));
+    free (newpaths);
+  }
+
+  {
+    pk_val exception;
+
+    /* The Poke hyperlinks facilities must be loaded before poke.pk and
+       the cmd subsystem.  This is done even if the hserver is
+       disabled.  */
+    if (pk_load (poke_compiler, "pk-hserver", &exception) != PK_OK
+        || exception != PK_NULL)
+      /* There's no default exception handler yet to handle `exception'.  */
+      pk_fatal ("unable to load the pk-hserver module");
+
+    /* Load poke.pk  */
+    if (pk_load (poke_compiler, "poke", &exception) != PK_OK
+        || exception != PK_NULL)
+      /* There's no default exception handler yet to handle `exception'.  */
+      pk_fatal ("unable to load the poke module");
+  }
 
   /* Set the values of a few global variables defined in poke.pk.  */
   {
@@ -638,13 +709,13 @@ initialize (int argc, char *argv[])
     pk_decl_set_val (poke_compiler, "pk_network_endian", network_endian);
 
     pk_decl_set_val (poke_compiler, "pk_auto_map_p",
-                     pk_make_int (poke_default_auto_map_p, 32));
+                     pk_make_int (poke_compiler, poke_default_auto_map_p, 32));
 
     pk_decl_set_val (poke_compiler, "pk_have_libnbd_p",
 #if defined HAVE_LIBNBD
-                     pk_make_int (1, 32)
+                     pk_make_int (poke_compiler, 1, 32)
 #else
-                     pk_make_int (0, 32)
+                     pk_make_int (poke_compiler, 0, 32)
 #endif
                      );
   }
@@ -652,6 +723,9 @@ initialize (int argc, char *argv[])
   /* Initialize the global map.  */
   pk_map_init ();
 
+  /* Initialize ios stuff.  */
+  pk_ios_init ();
+
   /* Initialize the command subsystem.  This should be done even if
      called non-interactively.  */
   pk_cmd_init ();
@@ -659,7 +733,6 @@ initialize (int argc, char *argv[])
 #ifdef HAVE_HSERVER
   poke_hserver_p = (poke_interactive_p
                     && pk_term_color_p ()
-                    && !poke_mi_p
                     && !poke_no_hserver_arg);
 
   /* Initialize and start the terminal hyperlinks server.  */
@@ -671,7 +744,7 @@ initialize (int argc, char *argv[])
 
   /* Set the value of the Poke variable pk_hserver_p.  */
   pk_decl_set_val (poke_compiler, "pk_hserver_p",
-                   pk_make_int (poke_hserver_p, 32));
+                   pk_make_int (poke_compiler, poke_hserver_p, 32));
 #endif
 }
 
@@ -700,27 +773,52 @@ initialize_user (void)
 
   /* If no ~/.pokerc file was found, acknowledge the XDG Base
      Directory Specification, as documented in
-     https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
-
-     If the environment variable XDG_CONFIG_HOME if defined, try to
-     load the file $XDG_CONFIG_HOME/pokerc.conf.
-
-     If the environment variable $XDG_CONFIG_DIRS is defined, try to
-     use all the : separated paths in that variable to find
-     pokerc.conf.  Else, try to load /etc/xdg/poke/pokerc.conf.  */
+     https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html */
   {
-    const char *xdg_config_home = getenv ("XDG_CONFIG_HOME");
-    const char *xdg_config_dirs = getenv ("XDG_CONFIG_DIRS");
-
-    if (xdg_config_home == NULL)
-      xdg_config_home = "";
+    char *xdg_config_dirs = NULL;
+    char *xdg_config_home = NULL;
 
+    /* If the environment variable $XDG_CONFIG_DIRS is defined, try to
+       use all the : separated paths in that variable to find
+       pokerc.conf.  Else, try to load /etc/xdg/poke/pokerc.conf.  */
+    xdg_config_dirs = getenv ("XDG_CONFIG_DIRS");
     if (xdg_config_dirs == NULL)
       xdg_config_dirs = "/etc/xdg";
 
-    char *config_path = pk_str_concat (xdg_config_dirs, ":", xdg_config_home, NULL);
-    pk_assert_alloc (config_path);
+    char *config_path = xstrdup (xdg_config_dirs);
 
+    /* If the environment variable XDG_CONFIG_HOME if defined, try to
+       load the file $XDG_CONFIG_HOME/pokerc.conf.  */
+    xdg_config_home = getenv ("XDG_CONFIG_HOME");
+    if (xdg_config_home == NULL && homedir != NULL)
+      {
+        /* If unset, the default value of $HOME/.config should be preferred to
+           the paths in $XDG_CONFIG_DIRS. From the standard:
+
+           A user-specific version of the configuration file may be created in
+           $XDG_CONFIG_HOME/subdir/filename, taking into account the default
+           value for $XDG_CONFIG_HOME if $XDG_CONFIG_HOME is not set. */
+        xdg_config_home = pk_str_concat (homedir, "/.config", NULL);
+        pk_assert_alloc (xdg_config_home);
+      }
+    else if (xdg_config_home != NULL)
+      {
+        /* strdup() the user-provided value, to free it below. */
+        xdg_config_home = strdup (xdg_config_home);
+        pk_assert_alloc (xdg_config_home);
+      }
+
+    if (xdg_config_home)
+      {
+        char *new_config_path = pk_str_concat (xdg_config_home, ":", config_path, NULL);
+
+        free (config_path);
+        free (xdg_config_home);
+        pk_assert_alloc (new_config_path);
+        config_path = new_config_path;
+      }
+
+    /* Ok now process config_path.  */
     char *dir = strtok (config_path, ":");
     do
       {
@@ -756,17 +854,18 @@ poke_handle_exception (pk_val exception)
   pk_val handler = pk_decl_val (poke_compiler, "pk_exception_handler");
 
   assert (handler != PK_NULL);
-  if (pk_call (poke_compiler, handler, NULL, 1, exception) == PK_ERROR)
-    assert (0);
+  if (pk_call (poke_compiler, handler, NULL, NULL, 1, exception)
+      == PK_ERROR)
+    PK_UNREACHABLE ();
 }
 
 void
 pk_fatal (const char *errmsg)
 {
   if (errmsg)
-    pk_printf ("fatal error: %s\n", errmsg);
-  pk_printf ("This is a bug. Please report it to %s\n",
-             PACKAGE_BUGREPORT);
+    fprintf (stderr, "fatal error: %s\n", errmsg);
+  fprintf (stderr, "This is a bug. Please report it to %s\n",
+           PACKAGE_BUGREPORT);
   abort ();
 }
 
@@ -791,13 +890,8 @@ main (int argc, char *argv[])
      for IO, etc etc */
   parse_args_2 (argc, argv);
 
-  /* Enter the REPL or MI.  */
-  if (poke_mi_p)
-    {
-      if (!pk_mi ())
-        poke_exit_code = EXIT_FAILURE;
-    }
-  else if (poke_interactive_p)
+  /* Enter the REPL.  */
+  if (poke_interactive_p)
     pk_repl ();
 
   /* Cleanup.  */
